/*
 * This file is part of Barista.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Barista is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Barista is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package fr.x9c.barista.api;

import java.util.Arrays;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;

import fr.x9c.cadmium.kernel.Block;
import fr.x9c.cadmium.kernel.Value;
import fr.x9c.cadmium.support.Helper;
import fr.x9c.cadmium.support.values.ToValue;

/**
 * This class represents a Java annotation.
 *
 * @author <a href="mailto:barista@x9c.fr">Xavier Clerc</a>
 * @version 1.0
 * @since 1.0
 */
public final class Annotation implements ToValue {

    /**
     * Parent class of all element value embedded inside annotations.
     */
    public static abstract class ElementValue implements ToValue {

        /**
         * Ensures that no element value could be defined outside this file.
         */
        private ElementValue() {
        } // end empty constructor

    } // end inner-class 'ElementValue'

    /**
     * Represents a boolean value embedded inside an annotation.
     */
    public static final class BooleanValue extends ElementValue {

        /** The underlying value. */
        private final boolean value;

        /**
         * Constructs a boolean value.
         * @param v value
         */
        public BooleanValue(final boolean v) {
            this.value = v;
        } // end constructor(boolean)

        /**
         * Returns the underlying value.
         * @return the underlying value
         */
        public boolean getValue() {
            return this.value;
        } // end method 'getValue()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(0, this.value ? Value.TRUE : Value.FALSE);
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "BooleanValue(" + this.value + ")";
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return this.value ? 1 : 0;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof BooleanValue) {
                final BooleanValue that = (BooleanValue) obj;
                return this.value == that.value;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'BooleanValue'

    /**
     * Represents a byte value embedded inside an annotation.
     */
    public static final class ByteValue extends ElementValue {

        /** The underlying value. */
        private final byte value;

        /**
         * Constructs a byte value.
         * @param v value
         */
        public ByteValue(final byte v) {
            this.value = v;
        } // end constructor(byte)

        /**
         * Returns the underlying value.
         * @return the underlying value
         */
        public byte getValue() {
            return this.value;
        } // end method 'getValue()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, Value.createFromLong(this.value));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "ByteValue(" + this.value + ")";
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return this.value;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof ByteValue) {
                final ByteValue that = (ByteValue) obj;
                return this.value == that.value;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'ByteValue'

    /**
     * Represents a char value embedded inside an annotation.
     */
    public static final class CharValue extends ElementValue {

        /** The underlying value. */
        private final char value;

        /**
         * Constructs a char value.
         * @param v value
         */
        public CharValue(final char v) {
            this.value = v;
        } // end constructor(char)

        /**
         * Returns the underlying value.
         * @return the underlying value
         */
        public char getValue() {
            return this.value;
        } // end method 'getValue()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(2, Value.createFromLong(this.value));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "CharValue(" + this.value + ")";
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return this.value;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof CharValue) {
                final CharValue that = (CharValue) obj;
                return this.value == that.value;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'CharValue'

    /**
     * Represents a double value embedded inside an annotation.
     */
    public static final class DoubleValue extends ElementValue {

        /** The underlying value. */
        private final double value;

        /**
         * Constructs a double value.
         * @param v value
         */
        public DoubleValue(final double v) {
            this.value = v;
        } // end constructor(double)

        /**
         * Returns the underlying value.
         * @return the underlying value
         */
        public double getValue() {
            return this.value;
        } // end method 'getValue()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(3, Helper.createFloat(this.value));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "DoubleValue(" + this.value + ")";
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return (int) this.value;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof DoubleValue) {
                final DoubleValue that = (DoubleValue) obj;
                return this.value == that.value;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'DoubleValue'

    /**
     * Represents a float value embedded inside an annotation.
     */
    public static final class FloatValue extends ElementValue {

        /** The underlying value. */
        private final float value;

        /**
         * Constructs a float value.
         * @param v value
         */
        public FloatValue(final float v) {
            this.value = v;
        } // end constructor(float)

        /**
         * Returns the underlying value.
         * @return the underlying value
         */
        public float getValue() {
            return this.value;
        } // end method 'getValue()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(4, Helper.createFloat(this.value));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "FloatValue(" + this.value + ")";
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return (int) this.value;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof FloatValue) {
                final FloatValue that = (FloatValue) obj;
                return this.value == that.value;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'FloatValue'

    /**
     * Represents an int value embedded inside an annotation.
     */
    public static final class IntValue extends ElementValue {

        /** The underlying value. */
        private final int value;

        /**
         * Constructs a int value.
         * @param v value
         */
        public IntValue(final int v) {
            this.value = v;
        } // end constructor(int)

        /**
         * Returns the underlying value.
         * @return the underlying value
         */
        public int getValue() {
            return this.value;
        } // end method 'getValue()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(5, Helper.createInt32(this.value));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "IntValue(" + this.value + ")";
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return this.value;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof IntValue) {
                final IntValue that = (IntValue) obj;
                return this.value == that.value;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'IntValue'

    /**
     * Represents a long value embedded inside an annotation.
     */
    public static final class LongValue extends ElementValue {

        /** The underlying value. */
        private final long value;

        /**
         * Constructs a long value.
         * @param v value
         */
        public LongValue(final long v) {
            this.value = v;
        } // end constructor(long)

        /**
         * Returns the underlying value.
         * @return the underlying value
         */
        public long getValue() {
            return this.value;
        } // end method 'getValue()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(6, Helper.createInt64(this.value));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "LongValue(" + this.value + ")";
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return (int) this.value;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof LongValue) {
                final LongValue that = (LongValue) obj;
                return this.value == that.value;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'LongValue'

    /**
     * Represents a short value embedded inside an annotation.
     */
    public static final class ShortValue extends ElementValue {

        /** The underlying value. */
        private final short value;

        /**
         * Constructs a short value.
         * @param v value
         */
        public ShortValue(final short v) {
            this.value = v;
        } // end constructor(short)

        /**
         * Returns the underlying value.
         * @return the underlying value
         */
        public short getValue() {
            return this.value;
        } // end method 'getValue()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(7, Value.createFromLong(this.value));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "ShortValue(" + this.value + ")";
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return this.value;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof ShortValue) {
                final ShortValue that = (ShortValue) obj;
                return this.value == that.value;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'ShortValue'

    /**
     * Represents a string value embedded inside an annotation.
     */
    public static final class StringValue extends ElementValue {

        /** The underlying value. */
        private final String value;

        /**
         * Constructs a string value.
         * @param v value - should not be <tt>null</tt>
         */
        public StringValue(final String v) {
            assert v != null : "null v";
            this.value = v;
        } // end constructor(String)

        /**
         * Returns the underlying value.
         * @return the underlying value
         */
        public String getValue() {
            return this.value;
        } // end method 'getValue()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(8, API.createUTF8(this.value));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "StringValue(" + this.value + ")";
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return this.value.hashCode();
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof StringValue) {
                final StringValue that = (StringValue) obj;
                return this.value.equals(that.value);
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'StringValue'

    /**
     * Represents an enum value embedded inside an annotation.
     */
    public static final class EnumValue extends ElementValue {

        /** The underlying class name. */
        private final String className;

        /** The underlying enum name. */
        private final String enumName;

        /**
         * Constructs an enum value.
         * @param cn class name - should not be <tt>null</tt>
         * @param en enum name - should not be <tt>null</tt>
         */
        public EnumValue(final String cn, final String en) {
            assert cn != null : "null cn";
            assert en != null : "null en";
            this.className = cn;
            this.enumName = en;
        } // end constructor(String, String)

        /**
         * Returns the class name.
         * @return the class name
         */
        public String getClassName() {
            return this.className;
        } // end method 'getClassName()'

        /**
         * Returns the enum name.
         * @return the enum name
         */
        public String getEnumName() {
            return this.enumName;
        } // end method 'getEnumName()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(9,
                                              API.createClassName(this.className),
                                              API.createFieldName(this.enumName));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "EnumValue(" + this.className + ", " + this.enumName + ")";
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return this.className.hashCode() + this.enumName.hashCode();
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof EnumValue) {
                final EnumValue that = (EnumValue) obj;
                return this.className.equals(that.className)
                    && this.enumName.equals(that.enumName);
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'EnumValue'

    /**
     * Represents a class value embedded inside an annotation.
     */
    public static final class ClassValue extends ElementValue {

        /** The underlying class name. */
        private final String className;

        /**
         * Constructs a class value.
         * @param cn class name - should not be <tt>null</tt>
         */
        public ClassValue(final String cn) {
            assert cn != null : "null cn";
            this.className = cn;
        } // end constructor(String)

        /**
         * Returns the class name.
         * @return the class name
         */
        public String getClassName() {
            return this.className;
        } // end method 'getClassName()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(10, API.createClassName(this.className));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "ClassValue(" + this.className + ")";
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return this.className.hashCode();
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof ClassValue) {
                final ClassValue that = (ClassValue) obj;
                return this.className.equals(that.className);
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'ClassValue'

    /**
     * Represents an annotation value embedded inside an annotation.
     */
    public static final class AnnotationValue extends ElementValue {

        /** The underlying value. */
        private final Annotation value;

        /**
         * Constructs an annotation value.
         * @param v value - should not be <tt>null</tt>
         */
        public AnnotationValue(final Annotation v) {
            assert v != null : "null v";
            this.value = v;
        } // end constructor(Annotation)

        /**
         * Returns the underlying value.
         * @return the underlying value
         */
        public Annotation getValue() {
            return this.value;
        } // end method 'getValue()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(11, this.value.toValue());
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "AnnotationValue(" + this.value + ")";
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return this.value.hashCode();
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof AnnotationValue) {
                final AnnotationValue that = (AnnotationValue) obj;
                return this.value.equals(that.value);
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'AnnotationValue'

    /**
     * Represents an Array value embedded inside an annotation.
     */
    public static final class ArrayValue extends ElementValue {

        /** The underlying value. */
        private final ElementValue[] value;

        /**
         * Constructs an array value.
         * @param v value - should not be <tt>null</tt>
         */
        public ArrayValue(final ElementValue[] v) {
            assert v != null : "null v";
            final int len = v.length;
            this.value = new ElementValue[len];
            System.arraycopy(v, 0, this.value, 0, len);
        } // end constructor(ElementValue[])

        /**
         * Returns the underlying value.
         * @return the underlying value
         */
        public ElementValue[] getValue() {
            final int len = this.value.length;
            final ElementValue[] res = new ElementValue[len];
            System.arraycopy(this.value, 0, res, 0, len);
            return res;
        } // end method 'getValue()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            Value res = Value.EMPTY_LIST;
            for (int i = this.value.length - 1; i >= 0; i--) {
                final Block cons = Block.createBlock(Block.TAG_CONS,
                                                     this.value[i].toValue(),
                                                     res);
                res = Value.createFromBlock(cons);
            } // end for
            final Block b = Block.createBlock(12, res);
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "ArrayValue(" + Arrays.toString(this.value) + ")";
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return Arrays.hashCode(this.value);
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof ArrayValue) {
                final ArrayValue that = (ArrayValue) obj;
                return Arrays.equals(this.value, that.value);
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'ArrayValue'

    /** Annotation name. */
    private final String name;

    /** Annotation element names. */
    private final List<String> names;

    /** Annotation element values. */
    private final List<? extends ElementValue> values;

    /**
     * Constructs an annotation.
     * @param nm annotation name - should not be <tt>null</tt>
     * @param n element names - should not be <tt>null</tt>,
     *                          should also have the same size as v
     * @param v element values - should not be <tt>null</tt>,
     *                           should also have the same size as n
     */
    public Annotation(final String nm,
                      final List<String> n,
                      final List<? extends ElementValue> v) {
        assert nm != null : "null nm";
        assert n != null : "null n";
        assert v != null : "null v";
        assert n.size() == v.size() : "n and v have different sizes";
        this.name = nm;
        this.names = new LinkedList<String>(n);
        this.values = new LinkedList<ElementValue>(v);
    } // end constructor(String, List<String>, List<? extends ElementValue>)

    /**
     * Returns the name of the annotation.
     * @return the name of the annotation
     */
    public String getName() {
        return this.name;
    } // end method 'getName()'

    /**
     * Returns the names of the annotations elements.
     * @return the names of the annotations elements
     */
    public List<String> getNames() {
        return new LinkedList<String>(this.names);
    } // end method 'getNames()'

    /**
     * Returns the values of the annotations elements.
     * @return the values of the annotations elements
     */
    public List<? extends ElementValue> getValues() {
        return new LinkedList<ElementValue>(this.values);
    } // end method 'getValues()'

    /**
     * {@inheritDoc}
     */
    public Value toValue() {
        Value list = Value.EMPTY_LIST;
        final ListIterator<String> itNames = this.names.listIterator(this.names.size());
        final ListIterator<? extends ElementValue> itValues = this.values.listIterator(this.values.size());
        while (itNames.hasPrevious() && itValues.hasPrevious()) {
            final Block b = Block.createBlock(0,
                                              API.createUTF8(itNames.previous()),
                                              itValues.previous().toValue());
            final Block cons = Block.createBlock(Block.TAG_CONS,
                                                 Value.createFromBlock(b),
                                                 list);
            list = Value.createFromBlock(cons);
        } // end while
        final Block res = Block.createBlock(0,
                                            API.createClassName(this.name),
                                            list);
        return Value.createFromBlock(res);
    } // end method 'toValue()'

    /**
     * Constructs an instance from a value.
     * @param v value to construct instance from - should not be <tt>null</tt>
     * @return the corresponding instance
     */
    static ElementValue elementFromValue(final Value v) {
        assert v != null : "null v";
        final Block val = v.asBlock();
        final ElementValue elem;
        switch (val.getTag()) {
        case 0:
            return new BooleanValue(val.get(0) == Value.TRUE);
        case 1:
            return new ByteValue((byte) val.get(0).asLong());
        case 2:
            return new CharValue((char) val.get(0).asLong());
        case 3:
            return new DoubleValue(val.get(0).asBlock().asDouble());
        case 4:
            return new FloatValue((float) val.get(0).asBlock().asDouble());
        case 5:
            return new IntValue(val.get(0).asBlock().asInt32());
        case 6:
            return new LongValue(val.get(0).asBlock().asInt64());
        case 7:
            return new ShortValue((short) val.get(0).asLong());
        case 8:
            return new StringValue(API.stringOfUTF8(val.get(0)));
        case 9:
            return new EnumValue(API.stringOfClassName(val.get(0)),
                                 API.stringOfFieldName(val.get(1)));
        case 10:
            return new ClassValue(API.stringOfClassName(val.get(0)));
        case 11:
            return new AnnotationValue(fromValue(val.get(0)));
        case 12:
            final List<Value> elements = API.decodeList(val.get(0));
            final int sz = elements.size();
            final ElementValue[] elementValues = new ElementValue[sz];
            elements.toArray(elementValues);
            return new ArrayValue(elementValues);
        default:
            assert false : "invalid tag";
            return null;
        } // end switch
    } // end method 'elementFromValue(Value)'

    /**
     * Constructs an instance from a value.
     * @param v value to construct instance from - should not be <tt>null</tt>
     * @return the corresponding instance
     */
    static Annotation fromValue(final Value v) {
        assert v != null : "null v";
        final Block b = v.asBlock();
        final String name = API.stringOfClassName(b.get(0));
        final List<Value> list = API.decodeList(b.get(1));
        final List<String> names = new LinkedList<String>();
        final List<ElementValue> values = new LinkedList<ElementValue>();
        for (Value e : list) {
            final Block bl = e.asBlock();
            names.add(API.stringOfUTF8(bl.get(0)));
            values.add(elementFromValue(bl.get(1)));
        } // end for
        return new Annotation(name, names, values);
    } // end method 'fromValue(Value)'

    /**
     * {@inheritDoc}
     */
    @Override
    public String toString() {
        final StringBuilder sb = new StringBuilder();
        sb.append("@");
        sb.append(this.name);
        sb.append("(");
        final Iterator<String> itNames = this.names.iterator();
        final Iterator<? extends ElementValue> itValues = this.values.iterator();
        boolean first = true;
        while (itNames.hasNext() && itValues.hasNext()) {
            if (first) {
                first = false;
            } else {
                sb.append(", ");
            } // end if/else
            sb.append(itNames.next());
            sb.append("=");
            sb.append(itValues.next().toString());
        } // end while
        sb.append(")");
        return sb.toString();
    } // end method 'toString()'

    /**
     * {@inheritDoc}
     */
    @Override
    public int hashCode() {
        return this.name.hashCode()
            + this.names.hashCode()
            + this.values.hashCode();
    } // end method 'hashCode()'

    /**
     * {@inheritDoc}
     */
    @Override
    public boolean equals(final Object obj) {
        if (obj instanceof Annotation) {
            final Annotation that = (Annotation) obj;
            return this.name.equals(that.name)
                && this.names.equals(that.names)
                && this.values.equals(that.values);
        } else {
            return false;
        } // end if/else
    } // end method 'equals(Object)'

} // end class 'Annotation'
