/*
 * This file is part of Barista.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Barista is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Barista is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package fr.x9c.barista.api;

import java.util.LinkedList;
import java.util.List;

import fr.x9c.cadmium.kernel.Block;
import fr.x9c.cadmium.kernel.Value;
import fr.x9c.cadmium.support.values.ToValue;

/**
 * This class represents a Java field.
 *
 * @author <a href="mailto:barista@x9c.fr">Xavier Clerc</a>
 * @version 1.0
 * @since 1.0
 */
public final class Field implements ToValue {

    /** Field flags. */
    private final List<AccessFlag> flags;

    /** Field name. */
    private final String name;

    /** Field descriptor. */
    private final Descriptor descriptor;

    /** Field attribute. */
    private final List<? extends Attribute> attributes;

    /**
     * Constructs a field.
     * @param f field flags - should not be <tt>null</tt>,
     *          should also be field flags
     * @param n field name - should not be <tt>null</tt>
     * @param d field descriptor - should not be <tt>null</tt>,
     *          should also be a non-void descriptor
     * @param a field attributes - should not be <tt>null</tt>,
     *          should also be field attributes
     * @throws BaristaException if the descriptor is not a valid descriptor
     * @throws BaristaException if one of the flags is not a valid flag
     */
    public Field(final List<AccessFlag> f,
                 final String n,
                 final Descriptor d,
                 final List<? extends Attribute> a)
        throws BaristaException {
        assert f != null : "null f";
        assert n != null : "null n";
        assert d != null : "null d";
        assert a != null : "null a";
        if (!AccessFlag.areFieldFlags(f)) {
            throw new BaristaException("invalid flag");
        } // end if
        if (!d.isFieldDescriptor()) {
            throw new BaristaException("invalid descriptor");
        } // end if
        if (!Attribute.areFieldAttributes(a)) {
            throw new BaristaException("invalid attribute");
        } // end if
        this.flags = new LinkedList<AccessFlag>(f);
        this.name = n;
        this.descriptor = d;
        this.attributes = new LinkedList<Attribute>(a);
    } // end constructor(List<AccessFlag>, String, Descriptor, List<? extends Attribute>)

    /**
     * Returns the field's flags.
     * @return the field's flags
     */
    public List<AccessFlag> getFlags() {
        return new LinkedList<AccessFlag>(this.flags);
    } // end method 'getFlags()'

    /**
     * Returns the field's name.
     * @return the field's name
     */
    public String getName() {
        return this.name;
    } // end method 'getName()'

    /**
     * Returns the field's descriptor.
     * @return the field's descriptor
     */
    public Descriptor getDescriptor() {
        return this.descriptor;
    } // end method 'getDescriptor()'

    /**
     * Returns the field's attributes.
     * @return the field's attributes
     */
    public List<? extends Attribute> getAttributes() {
        return new LinkedList<Attribute>(this.attributes);
    } // end method 'getAttributes()'

    /**
     * {@inheritDoc}
     */
    public Value toValue() {
        final Block b = Block.createBlock(0,
                                          API.encodeList(this.flags),
                                          API.createFieldName(this.name),
                                          this.descriptor.toValue(),
                                          API.encodeList(this.attributes));
        return Value.createFromBlock(b);
    } // end method 'toValue()'

    /**
     * Constructs an instance from a value.
     * @param v value to construct instance from - should not be <tt>null</tt>
     * @return the corresponding instance
     */
    static Field fromValue(final Value v) {
        assert v != null : "null v";
        final Block b = v.asBlock();
        final List<AccessFlag> flags = new LinkedList<AccessFlag>();
        for (Value f : API.decodeList(b.get(0))) {
            flags.add(AccessFlag.fromValue(f));
        } // end for
        final String name = API.stringOfFieldName(b.get(1));
        final Descriptor desc = Descriptor.fromValue(b.get(2));
        final List<Attribute> attrs = new LinkedList<Attribute>();
        for (Value a : API.decodeList(b.get(3))) {
            attrs.add(Attribute.fromValue(a));
        } // end for
        return new Field(flags, name, desc, attrs);
    } // end method 'fromValue(Value)'

    /**
     * {@inheritDoc}
     */
    @Override
    public String toString() {
        final StringBuilder sb = new StringBuilder();
        final String f = AccessFlag.listToString(this.flags);
        if (f.length() > 0) {
            sb.append(f);
            sb.append(" ");
        } // end if
        sb.append(this.descriptor.toString());
        sb.append(" ");
        sb.append(this.name);
        return sb.toString();
    } // end method 'toString()'

    /**
     * {@inheritDoc}
     */
    @Override
    public int hashCode() {
        return this.flags.hashCode()
            + this.name.hashCode()
            + this.descriptor.hashCode()
            + this.attributes.hashCode();
    } // end method 'hashCode()'

    /**
     * {@inheritDoc}
     */
    @Override
    public boolean equals(final Object obj) {
        if (obj instanceof Field) {
            final Field that = (Field) obj;
            return this.flags.equals(that.flags)
                && this.name.equals(that.name)
                && this.descriptor.equals(that.descriptor)
                && this.attributes.equals(that.attributes);
        } else {
            return false;
        } // end if/else
    } // end method 'equals(Object)'

} // end class 'Field'
