/*
 * This file is part of Barista.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Barista is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Barista is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package fr.x9c.barista.api;

import java.util.LinkedList;
import java.util.List;

import fr.x9c.cadmium.kernel.Block;
import fr.x9c.cadmium.kernel.Value;
import fr.x9c.cadmium.support.values.ToValue;

/**
 * This class represents a Java method.
 *
 * @author <a href="mailto:barista@x9c.fr">Xavier Clerc</a>
 * @version 1.0
 * @since 1.0
 */
public final class Method implements ToValue {

    /** This enum defines the different method kinds. */
    public enum Kind { REGULAR, CONSTRUCTOR, INITIALIZER }

    /** Method kind. */
    private final Kind kind;

    /** Method flags. */
    private final List<AccessFlag> flags;

    /** Method name. */
    private final String name;

    /** Method return type. */
    private final Descriptor returnType;

    /** Method parameter types. */
    private final List<? extends Descriptor> parameterTypes;

    /** Method attributes. */
    private final List<? extends Attribute> attributes;

    /** Whether initializer is <i>strictfp</i>. */
    private final boolean strict;

    /**
     * Constructs a regular method.
     * @param f method flags - should not be <tt>null</tt>,
     *          should also be method flags
     * @param n method name - should not be <tt>null</tt>
     * @param r method return type - should not be <tt>null</tt>
     * @param p method parameter types - should not be <tt>null</tt>,
     *          should also be non-void descriptors
     * @param a method attributes - should not be <tt>null</tt>,
     *          should also be method attributes
     * @throws BaristaException if one of the flags is not a valid flag
     * @throws BaristaException if one of the descriptors is not a valid descriptor
     * @throws BaristaException if one of the attributes is not a valid attribute
     */
    public Method(final List<AccessFlag> f,
                  final String n,
                  final Descriptor r,
                  final List<? extends Descriptor> p,
                  final List<? extends Attribute> a)
        throws BaristaException {
        assert f != null : "null f";
        assert n != null : "null n";
        assert r != null : "null r";
        assert p != null : "null p";
        assert a != null : "null a";
        if (!AccessFlag.areMethodFlags(f)) {
            throw new BaristaException("invalid flag");
        } // end if
        if (!Descriptor.areFieldDescriptors(p)) {
            throw new BaristaException("invalid descriptor");
        } // end if
        if (!Attribute.areMethodAttributes(a)) {
            throw new BaristaException("invalid attribute");
        } // end if
        this.kind = Kind.REGULAR;
        this.flags = new LinkedList<AccessFlag>(f);
        this.name = n;
        this.returnType = r;
        this.parameterTypes = new LinkedList<Descriptor>(p);
        this.attributes = new LinkedList<Attribute>(a);
        this.strict = false;
    } // end constructor(List<AccessFlag>, String, Descriptor, List<? extends Descriptor>, List<? extends Attribute>)

    /**
     * Constructs a constructor method.
     * @param f constructor flags - should not be <tt>null</tt>,
     *          should also be constructor flags
     * @param p constructor parameter types - should not be <tt>null</tt>,
     *          should also be non-void descriptors
     * @param a constructor attributes - should not be <tt>null</tt>,
     *          should also be method attributes
     * @throws BaristaException if one of the flags is not a valid flag
     * @throws BaristaException if one of the descriptors is not a valid descriptor
     * @throws BaristaException if one of the attributes is not a valid attribute
     */
    public Method(final List<AccessFlag> f,
                  final List<? extends Descriptor> p,
                  final List<? extends Attribute> a)
        throws BaristaException {
        assert f != null : "null f";
        assert p != null : "null p";
        assert a != null : "null a";
        if (!AccessFlag.areMethodFlags(f)) {
            throw new BaristaException("invalid flag");
        } // end if
        if (!Descriptor.areFieldDescriptors(p)) {
            throw new BaristaException("invalid descriptor");
        } // end if
        if (!Attribute.areMethodAttributes(a)) {
            throw new BaristaException("invalid attribute");
        } // end if
        this.kind = Kind.CONSTRUCTOR;
        this.flags = new LinkedList<AccessFlag>(f);
        this.name = null;
        this.returnType = null;
        this.parameterTypes = new LinkedList<Descriptor>(p);
        this.attributes = new LinkedList<Attribute>(a);
        this.strict = false;
    } // end constructor(List<AccessFlag>, List<? extends Descriptor>, List<? extends Attribute>)

    /**
     * Constructs an initializer method.
     * @param s whether initializer is <i>strictfp</i>
     * @param a initializer attributes - should not be <tt>null</tt>,
     *          should also be method attributes
     * @throws BaristaException if one of the attributes is not a valid attribute
     */
    public Method(final boolean s, final List<? extends Attribute> a)
        throws BaristaException {
        assert a != null : "null a";
        if (!Attribute.areMethodAttributes(a)) {
            throw new BaristaException("invalid attribute");
        } // end if
        this.kind = Kind.INITIALIZER;
        this.flags = null;
        this.name = null;
        this.returnType = null;
        this.parameterTypes = null;
        this.attributes = new LinkedList<Attribute>(a);
        this.strict = s;
    } // end constructor(boolean, List<? extends Attribute>)

    /**
     * Returns the method kind.
     * @return the method kind
     */
    public Kind getKind() {
        return this.kind;
    } // end method 'getKind()'

    /**
     * Returns the method flags.
     * @return the method flags
     */
    public List<AccessFlag> getFlags() {
        return new LinkedList<AccessFlag>(this.flags);
    } // end method 'getFlags()'

    /**
     * Returns the method name.
     * @return the method name
     */
    public String getName() {
        return this.name;
    } // end method 'getName()'

    /**
     * Returns the method return type.
     * @return the method return type
     */
    public Descriptor getReturnType() {
        return this.returnType;
    } // end method 'getReturnType()'

    /**
     * Returns the method parameter types.
     * @return the method parameter types
     */
    public List<? extends Descriptor> getParameterTypes() {
        return new LinkedList<Descriptor>(this.parameterTypes);
    } // end method 'getParameterTypes()'

    /**
     * Returns the method attributes.
     * @return the method attributes
     */
    public List<? extends Attribute> getAttribute() {
        return new LinkedList<Attribute>(this.attributes);
    } // end method 'getAttribute()'

    /**
     * Tests whether the method is an initializer that is <i>strictfp</i>.
     * @return <tt>true</tt> if the method is an initializer that is <i>strictfp</i>,
     *         <tt>false</tt> otherwise
     */
    public boolean isStrict() {
        return this.strict;
    } // end method 'isStrict()'

    /**
     * {@inheritDoc}
     */
    public Value toValue() {
        switch (this.kind) {
        case REGULAR:
            final Block desc = Block.createBlock(0,
                                                 API.encodeList(this.parameterTypes),
                                                 this.returnType.toValue());
            final Block b1 = Block.createBlock(0,
                                               API.encodeList(this.flags),
                                               API.createMethodName(this.name),
                                               Value.createFromBlock(desc),
                                               API.encodeList(this.attributes));
            return Value.createFromBlock(b1);
        case CONSTRUCTOR:
            final Block b2 = Block.createBlock(1,
                                               API.encodeList(this.flags),
                                               API.encodeList(this.parameterTypes),
                                               API.encodeList(this.attributes));
            return Value.createFromBlock(b2);
        case INITIALIZER:
            final Block b3 = Block.createBlock(2,
                                               this.strict ? Value.TRUE : Value.FALSE,
                                               API.encodeList(this.attributes));
            return Value.createFromBlock(b3);
        default:
            assert false : "invalid kind";
            return null;
        } // end switch
    } // end method 'toValue()'

    /**
     * Constructs an instance from a value.
     * @param v value to construct instance from - should not be <tt>null</tt>
     * @return the corresponding instance
     */
    static Method fromValue(final Value v) {
        assert v != null : "null v";
        final Block b = v.asBlock();
        switch (b.getTag()) {
        case 0:
            final List<AccessFlag> f1 = new LinkedList<AccessFlag>();
            for (Value f : API.decodeList(b.get(0))) {
                f1.add(AccessFlag.fromValue(f));
            } // end for
            final Descriptor r1 = Descriptor.fromValue(b.get(2).asBlock().get(1));
            final List<Descriptor> p1 = new LinkedList<Descriptor>();
            for (Value d : API.decodeList(b.get(2).asBlock().get(0))) {
                p1.add(Descriptor.fromValue(d));
            } // end for
            final List<Attribute> a1 = new LinkedList<Attribute>();
            for (Value a : API.decodeList(b.get(3))) {
                a1.add(Attribute.fromValue(a));
            } // end for
            return new Method(f1, API.stringOfMethodName(b.get(1)), r1, p1, a1);
        case 1:
            final List<AccessFlag> f2 = new LinkedList<AccessFlag>();
            for (Value f : API.decodeList(b.get(0))) {
                f2.add(AccessFlag.fromValue(f));
            } // end for
            final List<Descriptor> p2 = new LinkedList<Descriptor>();
            for (Value d : API.decodeList(b.get(1))) {
                p2.add(Descriptor.fromValue(d));
            } // end for
            final List<Attribute> a2 = new LinkedList<Attribute>();
            for (Value a : API.decodeList(b.get(2))) {
                a2.add(Attribute.fromValue(a));
            } // end for
            return new Method(f2, p2, a2);
        case 2:
            final List<Attribute> a3 = new LinkedList<Attribute>();
            for (Value a : API.decodeList(b.get(1))) {
                a3.add(Attribute.fromValue(a));
            } // end for
            return new Method(b.get(0) == Value.TRUE, a3);
        default:
            assert false : "invalid tag";
            return null;
        } // end switch
    } // end method 'fromValue(Value)'

    /**
     * {@inheritDoc}
     */
    @Override
    public String toString() {
        final StringBuilder sb = new StringBuilder();
        final String f = this.flags != null
            ? AccessFlag.listToString(this.flags)
            : "";
        switch (this.kind) {
        case REGULAR:
            if (f.length() > 0) {
                sb.append(f);
                sb.append(" ");
            } // end if
            sb.append(this.returnType.toString());
            sb.append(" ");
            sb.append(this.name);
            sb.append("(");
            sb.append(Descriptor.listToString(this.parameterTypes));
            sb.append(")");
            return sb.toString();
        case CONSTRUCTOR:
            if (f.length() > 0) {
                sb.append(f);
                sb.append(" ");
            } // end if
            sb.append("<init>");
            sb.append("(");
            sb.append(Descriptor.listToString(this.parameterTypes));
            sb.append(")");
            return sb.toString();
        case INITIALIZER:
            if (this.strict) {
                sb.append("strictfp ");
            } // end if
            sb.append("<clinit>()");
            return sb.toString();
        default:
            assert false : "invalid kind";
            return null;
        } // end switch
    } // end method 'toString()'

    /**
     * {@inheritDoc}
     */
    @Override
    public int hashCode() {
        return this.kind.hashCode()
            + (this.flags != null ? this.flags.hashCode() : 0)
            + (this.name != null ? this.name.hashCode() : 0)
            + (this.returnType != null ? this.returnType.hashCode() : 0)
            + (this.parameterTypes != null ? this.parameterTypes.hashCode() : 0)
            + (this.attributes.hashCode())
            + (this.strict ? 1 : 0);
    } // end method 'hashCode()'

    /**
     * {@inheritDoc}
     */
    @Override
    public boolean equals(final Object obj) {
        if (obj instanceof Method) {
            final Method that = (Method) obj;
            return this.kind.equals(that.kind)
                && API.nullOrEquals(this.flags, that.flags)
                && API.nullOrEquals(this.name, that.name)
                && API.nullOrEquals(this.returnType, that.returnType)
                && API.nullOrEquals(this.parameterTypes, that.parameterTypes)
                && this.attributes.equals(that.attributes)
                && (this.strict == that.strict);
        } else {
            return false;
        } // end if/else
    } // end method 'equals(Object)'

} // end class 'Method'
