/*
 * This file is part of Barista.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Barista is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Barista is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package fr.x9c.barista.api;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;

import fr.x9c.cadmium.kernel.Block;
import fr.x9c.cadmium.kernel.Channel;
import fr.x9c.cadmium.kernel.Custom;
import fr.x9c.cadmium.kernel.Value;

/**
 * This class provides methods for class pretty-printing.
 *
 * @author <a href="mailto:barista@x9c.fr">Xavier Clerc</a>
 * @version 1.0
 * @since 1.0
 */
public final class Printer {

    /**
     * No instance of this class.
     */
    private Printer() {
    } // end empty constructor

    /**
     * Calls the callback to print a class.
     * @param name class name - should not be <tt>null</tt>
     * @param cp classpath for class loading - should not be <tt>null</tt>
     * @param out where to output class description - should not be <tt>null</tt>
     * @throws BaristaException if an errors occurs while printing
     */
    private static void callPrint(final String name,
                                  final ClassPath cp,
                                  final OutputStream out)
        throws BaristaException {
        assert name != null : "null name";
        assert cp != null : "null cp";
        assert out != null : "null out";
        final Value n = API.createUTF8(name);
        final Value c = cp.toValue();
        final Block b = Block.createCustom(Custom.CHANNEL_SIZE,
                                           Custom.CHANNEL_OPS);
        b.setCustom(new Channel(out));
        final Value o = Value.createFromBlock(b);
        API.callback("fr.x9c.barista.api.Printer.print", n, c, o);
    } // end method 'callPrint(String, ClassPath, OutputStream)'

    /**
     * Pretty-prints a class.
     * @param name class name - should not be <tt>null</tt>
     * @param cp classpath for class loading - should not be <tt>null</tt>
     * @param out where to output class description - should not be <tt>null</tt>
     * @throws BaristaException if an errors occurs while printing
     */
    public static void print(final String name,
                             final ClassPath cp,
                             final OutputStream out)
        throws BaristaException {
        callPrint(name, cp, out);
    } // end method 'print(String, ClassPath, OutputStream)'

    /**
     * Pretty-prints a class.
     * @param name class name - should not be <tt>null</tt>
     * @param cp classpath for class loading - should not be <tt>null</tt>
     * @param out where to output class description - should not be <tt>null</tt>
     * @throws BaristaException if an errors occurs while printing
     */
    public static void print(final String name,
                             final ClassPath cp,
                             final Appendable out)
        throws BaristaException {
        try {
            final ByteArrayOutputStream baos = new ByteArrayOutputStream();
            callPrint(name, cp, baos);
            out.append(baos.toString("UTF-8"));
        } catch (final UnsupportedEncodingException uee) {
            throw new BaristaException("UTF8 unsupported by Java platform");
        } catch (final IOException ioe) {
            throw new BaristaException("I/O error", ioe);
        } // and try/catch
    } // end method 'print(String, ClassPath, Appendable)'

    /**
     * Pretty-prints a class.
     * @param name class name - should not be <tt>null</tt>
     * @param cp classpath for class loading - should not be <tt>null</tt>
     * @return a string representation of the class
     * @throws BaristaException if an errors occurs while printing
     */
    public static String print(final String name, final ClassPath cp)
        throws BaristaException {
        try {
            final ByteArrayOutputStream baos = new ByteArrayOutputStream();
            callPrint(name, cp, baos);
            return baos.toString("UTF-8");
        } catch (java.io.UnsupportedEncodingException uee) {
            throw new BaristaException("UTF8 unsupported by Java platform");
        } // and try/catch
    } // end method 'print(String, ClassPath)'
    
} // end class 'Printer'
