(*
 * This file is part of Barista.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Barista is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Barista is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)

open Utils
open Consts


(* Low-level form *)

type info_element_value =
  | Primitive of [ `Boolean | `Byte | `Char | `Double | `Float | `Int | `Long | `Short] * u2
  | String of u2
  | Enum of u2 * u2
  | Class of u2
  | Annotation of info
  | Array of u2 * (info_element_value array)
and info = {
    type_index : u2;
    num_element_value_pairs : u2;
    element_value_pairs : (u2 * info_element_value) array;
  }


(* Exception *)

type error =
  | Invalid_tag
  | Inconsistent_primitive_value
  | Invalid_string_value
  | Invalid_enum_value
  | Invalid_class_value
  | Invalid_annotation_type_value
  | Invalid_element_name
  | Invalid_list_length

exception Exception of error

let fail e = raise (Exception e)

let string_of_error = function
  | Invalid_tag -> "invalid tag"
  | Inconsistent_primitive_value -> "inconsistent primitive value"
  | Invalid_string_value -> "invalid string value"
  | Invalid_enum_value -> "invalid enum value"
  | Invalid_class_value -> "invalid class value"
  | Invalid_annotation_type_value -> "invalid annotation type value"
  | Invalid_element_name -> "invalid element name"
  | Invalid_list_length -> "invalid list length"

let () =
  Printexc.register_printer
    (function
      | Exception e -> Some (string_of_error e)
      | _ -> None)


(* I/O functions *)

let rec read_info_element_value st =
  let tag = uchar_of_code (InputStream.read_u1 st) in
  if uchar_eq tag capital_b then
    let index = InputStream.read_u2 st in
    Primitive (`Byte, index)
  else if uchar_eq tag capital_c then
    let index = InputStream.read_u2 st in
    Primitive (`Char, index)
  else if uchar_eq tag capital_d then
    let index = InputStream.read_u2 st in
    Primitive (`Double, index)
  else if uchar_eq tag capital_f then
    let index = InputStream.read_u2 st in
    Primitive (`Float, index)
  else if uchar_eq tag capital_i then
    let index = InputStream.read_u2 st in
    Primitive (`Int, index)
  else if uchar_eq tag capital_j then
    let index = InputStream.read_u2 st in
    Primitive (`Long, index)
  else if uchar_eq tag capital_s then
    let index = InputStream.read_u2 st in
    Primitive (`Short, index)
  else if uchar_eq tag capital_z then
    let index = InputStream.read_u2 st in
    Primitive (`Boolean, index)
  else if uchar_eq tag small_s then
    let index = InputStream.read_u2 st in
    String index
  else if uchar_eq tag small_e then
    let type_name_index = InputStream.read_u2 st in
    let const_name_index = InputStream.read_u2 st in
    Enum (type_name_index, const_name_index)
  else if uchar_eq tag small_c then
    let class_info_index = InputStream.read_u2 st in
    Class class_info_index
  else if uchar_eq tag at_character then
    let annot = read_info st in
    Annotation annot
  else if uchar_eq tag opening_square_bracket then
    let num_values = InputStream.read_u2 st in
    let values = Array.init num_values (fun _ -> read_info_element_value st) in
    Array (num_values, values)
  else
    fail Invalid_tag
and read_info st =
  let type_idx = InputStream.read_u2 st in
  let nb = InputStream.read_u2 st in
  let evp = Array.init nb (fun _ ->
    let name_index = InputStream.read_u2 st in
    let el_vl = read_info_element_value st in
    (name_index, el_vl)) in { type_index = type_idx;
                              num_element_value_pairs = nb;
                              element_value_pairs = evp; }

let rec write_info_element_value st i =
  match i with
  | Primitive (jt, idx) ->
      let ch = match jt with
      | `Boolean -> capital_z
      | `Byte -> capital_b
      | `Char -> capital_c
      | `Double -> capital_d
      | `Float -> capital_f
      | `Int -> capital_i
      | `Long -> capital_l
      | `Short -> capital_s in
      OutputStream.write_u1 st (code_of_uchar ch);
      OutputStream.write_u2 st idx
  | String idx ->
      OutputStream.write_u1 st (code_of_uchar small_s);
      OutputStream.write_u2 st idx
  | Enum (n, v) ->
      OutputStream.write_u1 st (code_of_uchar small_e);
      OutputStream.write_u2 st n;
      OutputStream.write_u2 st v
  | Class idx ->
      OutputStream.write_u1 st (code_of_uchar small_c);
      OutputStream.write_u2 st idx
  | Annotation i ->
      OutputStream.write_u1 st (code_of_uchar at_character);
      write_info st i
  | Array (len, arr) ->
      OutputStream.write_u1 st (code_of_uchar opening_square_bracket);
      OutputStream.write_u2 st len;
      Array.iter (fun x -> write_info_element_value st x) arr
and write_info st i =
  OutputStream.write_u2 st i.type_index;
  OutputStream.write_u2 st i.num_element_value_pairs;
  Array.iter (fun (idx, iev) ->
    OutputStream.write_u2 st idx;
    write_info_element_value st iev)
    i.element_value_pairs


(* High-level form *)

type element_value =
  | Boolean_value of bool
  | Byte_value of int
  | Char_value of uchar
  | Double_value of float
  | Float_value of float
  | Int_value of int32
  | Long_value of int64
  | Short_value of int
  | String_value of utf8
  | Enum_value of Name.for_class * Name.for_field
  | Class_value of Name.for_class
  | Annotation_value of t
  | Array_value of element_value list
and t = Name.for_class * ((utf8 * element_value) list)


(* Conversion functions *)

let rec decode_element_value pool i =
  match i with
  | Primitive (d, idx) ->
      (match (ConstantPool.get_entry pool idx), d with
      | (ConstantPool.Integer v), `Boolean ->
          Boolean_value (v <> 0l)
      | (ConstantPool.Integer v), `Byte ->
          Byte_value (Int32.to_int v)
      | (ConstantPool.Integer v), `Char ->
          Char_value (uchar_of_code (Int32.to_int v))
      | (ConstantPool.Double (hi, lo)), `Double ->
          let d = Int64.logor
              (Int64.shift_left (Int64.of_int32 hi) 32)
              (Int64.of_int32 lo) in
          Double_value (Int64.float_of_bits d)
      | (ConstantPool.Float f), `Float ->
          Float_value (Int32.float_of_bits f)
      | (ConstantPool.Integer v), `Int ->
          Int_value v
      | (ConstantPool.Long (hi, lo)), `Long ->
          let v = Int64.logor
              (Int64.shift_left (Int64.of_int32 hi) 32)
              (Int64.of_int32 lo) in
          Long_value v
      | (ConstantPool.Integer v), `Short ->
          Short_value (Int32.to_int v)
      | _ -> fail Inconsistent_primitive_value)
  | String idx ->
      (match ConstantPool.get_entry pool idx with
      | ConstantPool.UTF8 n -> String_value n
      | _ -> fail Invalid_string_value)
  | Enum (name_idx, value_idx) ->
      (match (ConstantPool.get_entry pool name_idx),
        (ConstantPool.get_entry pool value_idx) with
      | (ConstantPool.UTF8 n), (ConstantPool.UTF8 v) ->
          (match Descriptor.field_of_utf8 n with
          | `Class cn ->
              Enum_value (cn, (Name.make_for_field v))
          | _ -> fail Invalid_enum_value)
      | _ -> fail Invalid_enum_value)
  | Class idx ->
      (match ConstantPool.get_entry pool idx with
      | ConstantPool.UTF8 n ->
          (match Descriptor.java_type_of_utf8 n with
          | `Class x -> Class_value x
          | _ -> fail Invalid_class_value)
      | _ -> fail Invalid_class_value)
  | Annotation i ->
      Annotation_value (decode pool i)
  | Array (_, arr) ->
      Array_value (List.map (decode_element_value pool) (Array.to_list arr))
and decode pool i =
    let type_desc = match ConstantPool.get_entry pool i.type_index with
    | ConstantPool.UTF8 n ->
        (match Descriptor.java_type_of_utf8 n with
        | `Class c -> c
        | _ -> fail Invalid_annotation_type_value)
    | _ -> fail Invalid_annotation_type_value in
    let pairs = List.map (fun (idx, iev) ->
      let name = match ConstantPool.get_entry pool idx with
      | ConstantPool.UTF8 n -> n
      | _ -> fail Invalid_element_name in
      (name, (decode_element_value pool iev)))
        (Array.to_list i.element_value_pairs) in
    (type_desc, pairs)

let rec encode_element_value pool e =
  match e with
  | Boolean_value v ->
      let idx = ConstantPool.add_integer pool (if v then 1l else 0l) in
      Primitive (`Boolean, idx)
  | Byte_value v ->
      let idx = ConstantPool.add_integer pool (Int32.of_int v) in
      Primitive (`Byte, idx)
  | Char_value v ->
      let idx = ConstantPool.add_integer pool (Int32.of_int (code_of_uchar v)) in
      Primitive (`Char, idx)
  | Double_value v ->
      let idx = ConstantPool.add_double pool v in
      Primitive (`Double, idx)
  | Float_value v ->
      let idx = ConstantPool.add_float pool v in
      Primitive (`Float, idx)
  | Int_value v ->
      let idx = ConstantPool.add_integer pool v in
      Primitive (`Int, idx)
  | Long_value v ->
      let idx = ConstantPool.add_long pool v in
      Primitive (`Long, idx)
  | Short_value v ->
      let idx = ConstantPool.add_integer pool (Int32.of_int v) in
      Primitive (`Short, idx)
  | String_value v ->
      let idx = ConstantPool.add_string pool v in
      String idx
  | Enum_value (n, v) ->
      let n' = (Descriptor.utf8_of_field (`Class n)) in
      let n_idx = ConstantPool.add_utf8 pool n' in
      let v_idx = ConstantPool.add_utf8 pool (Name.utf8_for_field v) in
      Enum (n_idx, v_idx)
  | Class_value n ->
      let desc = Descriptor.utf8_of_java_type (`Class n) in
      let idx = ConstantPool.add_utf8 pool desc in
      Class idx
  | Annotation_value a ->
      Annotation (encode pool a)
  | Array_value l ->
      let arr = Array.of_list (List.map (encode_element_value pool) l) in
      Array ((Array.length arr), arr)
and encode pool e =
  let checked_length l =
    let res = List.length l in
    if res < 65536 then
      res
    else
      fail Invalid_list_length in
  let (field, pairs) = e in
  let field_value = Descriptor.utf8_of_field (`Class field) in
  let field_index = ConstantPool.add_utf8 pool field_value in
  let pairs' = List.map (fun (n, v) ->
    let n_idx = ConstantPool.add_utf8 pool n in
    (n_idx, (encode_element_value pool v))) pairs in
  { type_index = field_index;
    num_element_value_pairs = checked_length pairs;
    element_value_pairs = Array.of_list pairs'; }
