(*
 * This file is part of Barista.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Barista is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Barista is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)

(** This module defines annotations in both low- and high-level forms.
   It also provides conversion functions between levels as well as i/o
   functions for low-level. *)


(** {6 Low-level form} *)

type info_element_value =
  | Primitive of [ `Boolean | `Byte | `Char | `Double | `Float | `Int | `Long | `Short] * Utils.u2 (** primitive value (given by type and index of value in constant pool) *)
  | String of Utils.u2 (** string value (given by index of value in constant pool) *)
  | Enum of Utils.u2 * Utils.u2 (** enum value (given by indexes of class and identifier) *)
  | Class of Utils.u2 (** class value (given by index of value in constant pool) *)
  | Annotation of info (** embedded annotation *)
  | Array of Utils.u2 * (info_element_value array) (** array of values *)
and info = {
    type_index : Utils.u2;
    num_element_value_pairs : Utils.u2;
    element_value_pairs : (Utils.u2 * info_element_value) array;
  }
(** Represents an annotation as defined in the class file format specification. *)


(** {6 Exception} *)

type error =
  | Invalid_tag
  | Inconsistent_primitive_value
  | Invalid_string_value
  | Invalid_enum_value
  | Invalid_class_value
  | Invalid_annotation_type_value
  | Invalid_element_name
  | Invalid_list_length

exception Exception of error
(** Exception to be raised when a function of this module fails.  *)

val string_of_error : error -> string
(** Converts the passed error into a string. *)


(** {6 I/O functions} *)

val read_info_element_value : InputStream.t -> info_element_value
(** [read_info_element_value st] reads an element value from [st].
   Raises [Exception] if [st] does not contain a valid element value.
   Raises [InputStream.Exception] if an i/o error occurs. *)

val write_info_element_value : OutputStream.t -> info_element_value -> unit
(** [write_info_element_value st e] writes element value [e] onto [st].
   Raises [Exception] if [e] is not a valid element value.
   Raises [OutputStream.Exception] if an i/o error occurs. *)

val read_info : InputStream.t -> info
(** [read_info st] reads an annotation from [st].
   Raises [Exception] if [st] does not contain a valid annotation.
   Raises [InputStream.Exception] if an i/o error occurs. *)

val write_info : OutputStream.t -> info -> unit
(** [write_info st a] writes annotation [a] onto [st].
   Raises [Exception] if [a] is not a valid annotation.
   Raises [OutputStream.Exception] if an i/o error occurs. *)


(** {6 High-level form} *)

type element_value =
  | Boolean_value of bool (** boolean value *)
  | Byte_value of int (** byte value *)
  | Char_value of Utils.uchar (** char value *)
  | Double_value of float (** double value *)
  | Float_value of float (** float value *)
  | Int_value of int32 (** int value *)
  | Long_value of int64 (** long value *)
  | Short_value of int (** short value *)
  | String_value of Utils.utf8 (** string value *)
  | Enum_value of Name.for_class * Name.for_field (** enum value (given by class and identifier) *)
  | Class_value of Name.for_class (** class value *)
  | Annotation_value of t (** embedded annotation *)
  | Array_value of element_value list (** array of values *)
and t = Name.for_class * ((Utils.utf8 * element_value) list)
(** Represents an annotation in high-level form. *)


(** {6 Conversion functions} *)

val decode_element_value : ConstantPool.t -> info_element_value -> element_value
(** Converts from a low-level into a high-level form according to passed pool.
   Raises [Exception] if an error occurs during conversion. *)

val decode : ConstantPool.t -> info -> t
(** Converts from a low-level into a high-level form according to passed pool.
   Raises [Exception] if an error occurs during conversion. *)

val encode_element_value : ConstantPool.extendable -> element_value -> info_element_value
(** Converts from a high-level into a low-level form, using passed extendable
   pool.
   Raises [Exception] if an error occurs during conversion. *)

val encode : ConstantPool.extendable -> t -> info
(** Converts from a high-level into a low-level form, using passed extendable
   pool.
   Raises [Exception] if an error occurs during conversion. *)
