(*
 * This file is part of Barista.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Barista is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Barista is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)

open Utils
open Consts


(* Low-level form *)

type info = {
    name_index : u2;
    length : u4;
    data : string;
  }


(* Exception *)

type error =
  | Invalid_code_attribute
  | Invalid_code_length
  | Defined_twice of string
  | Invalid_field_attribute
  | Invalid_method_attribute
  | Invalid_class_attribute
  | Invalid_constant_value
  | Invalid_enclosing_method
  | Invalid_local_variable_table
  | Invalid_local_variable_type_table
  | Invalid_list_length
  | Invalid_attribute_name_value
  | Invalid_exception_name
  | Invalid_exception
  | Invalid_inner_class
  | Invalid_outer_class
  | Invalid_signature
  | Invalid_source_file
  | Invalid_stack_map_frame
  | Invalid_stack_map_verification_type

exception Exception of error

let fail e = raise (Exception e)

let string_of_error = function
  | Invalid_code_attribute -> "invalid code attribute"
  | Invalid_code_length -> "invalid code length"
  | Defined_twice id -> id ^ " attribute defined twice"
  | Invalid_field_attribute -> "invalid field attribute"
  | Invalid_method_attribute -> "invalid method attribute"
  | Invalid_class_attribute -> "invalid class attribute"
  | Invalid_constant_value -> "invalid constant value"
  | Invalid_enclosing_method -> "invalid enclosing method"
  | Invalid_local_variable_table -> "invalid local variable table"
  | Invalid_local_variable_type_table -> "invalid local variable type table"
  | Invalid_list_length -> "invalid list length"
  | Invalid_attribute_name_value -> "invalid attribute name value"
  | Invalid_exception_name -> "invalid exception name"
  | Invalid_exception -> "invalid exception"
  | Invalid_inner_class -> "invalid inner class"
  | Invalid_outer_class -> "invalid outer class"
  | Invalid_signature -> "invalid signature"
  | Invalid_source_file -> "invalid source file"
  | Invalid_stack_map_frame -> "invalid stack map frame"
  | Invalid_stack_map_verification_type -> "invalid stack map verification type"

let () =
  Printexc.register_printer
    (function
      | Exception e -> Some (string_of_error e)
      | _ -> None)


(* I/O functions *)

let read_info st =
  let name = InputStream.read_u2 st in
  let len = InputStream.read_u4 st in
  if len > (Int64.of_int max_int) then
    raise (InputStream.Exception InputStream.Data_is_too_large)
  else
    let dat = InputStream.read_bytes st (Int64.to_int len) in
    { name_index = name;
      length = len;
      data = dat; }

let write_info st i =
  OutputStream.write_u2 st i.name_index;
  OutputStream.write_u4 st i.length;
  OutputStream.write_bytes st i.data


(* High-level form *)

type constant_value =
  | Long_value of int64
  | Float_value of float
  | Double_value of float
  | Boolean_value of bool
  | Byte_value of int
  | Character_value of int
  | Short_value of int
  | Integer_value of int32
  | String_value of utf8

type verification_type_info =
  | Top_variable_info
  | Integer_variable_info
  | Float_variable_info
  | Long_variable_info
  | Double_variable_info
  | Null_variable_info
  | Uninitialized_this_variable_info
  | Object_variable_info of [`Class_or_interface of Name.for_class | `Array_type of Descriptor.array_type]
  | Uninitialized_variable_info of u2

type stack_map_frame =
  | Same_frame of u2
  | Same_locals_1_stack_item_frame of u2 * verification_type_info
  | Chop_1_frame of u2
  | Chop_2_frame of u2
  | Chop_3_frame of u2
  | Append_1_frame of u2 * verification_type_info
  | Append_2_frame of u2 * verification_type_info * verification_type_info
  | Append_3_frame of u2 * verification_type_info * verification_type_info * verification_type_info
  | Full_frame of u2 * (verification_type_info list) * (verification_type_info list)

type t =
  [ `ConstantValue of constant_value
  | `Code of code_value
  | `Exceptions of Name.for_class list
  | `InnerClasses of ((Name.for_class option) * (Name.for_class option) * (utf8 option) * (AccessFlag.for_inner_class list)) list
  | `EnclosingMethod of Name.for_class * ((Name.for_method * Descriptor.for_method) option)
  | `Synthetic
  | `Signature of signature
  | `SourceFile of utf8
  | `SourceDebugExtension of utf8
  | `LineNumberTable of (u2 * u2) list
  | `LocalVariableTable of (u2 * u2 * utf8 * Descriptor.for_field * u2) list
  | `LocalVariableTypeTable of (u2 * u2 * utf8 * Signature.field_type_signature * u2) list
  | `Deprecated
  | `RuntimeVisibleAnnotations of Annotation.t list
  | `RuntimeInvisibleAnnotations of Annotation.t list
  | `RuntimeVisibleParameterAnnotations of Annotation.t list list
  | `RuntimeInvisibleParameterAnnotations of Annotation.t list list
  | `AnnotationDefault of Annotation.element_value
  | `StackMapTable of stack_map_frame list
  | `Unknown of utf8 * string ]
and signature =
  [ `Class of Signature.class_signature
  | `Method of Signature.method_signature
  | `Field of Signature.field_type_signature ]
and code_attribute = [
  | `LineNumberTable of (Utils.u2 * Utils.u2) list
  | `LocalVariableTable of (Utils.u2 * Utils.u2 * Utils.utf8 * Descriptor.for_field * Utils.u2) list
  | `LocalVariableTypeTable of (Utils.u2 * Utils.u2 * Utils.utf8 * Signature.field_type_signature * Utils.u2) list
  | `StackMapTable of stack_map_frame list
  | `Unknown of Utils.utf8 * string ]
and code_value = {
    max_stack : u2;
    max_locals : u2;
    code : Instruction.t list;
    exception_table : (u2 * u2 * u2 * (Name.for_class option)) list;
    attributes : code_attribute list;
  }

type for_field =
  [ `ConstantValue of constant_value
  | `Synthetic
  | `Signature of [`Field of Signature.field_type_signature]
  | `Deprecated
  | `RuntimeVisibleAnnotations of Annotation.t list
  | `RuntimeInvisibleAnnotations of Annotation.t list
  | `Unknown of utf8 * string ]

type for_method =
  [ `Code of code_value
  | `Exceptions of Name.for_class list
  | `Synthetic
  | `Signature of [`Method of Signature.method_signature]
  | `Deprecated
  | `RuntimeVisibleAnnotations of Annotation.t list
  | `RuntimeInvisibleAnnotations of Annotation.t list
  | `RuntimeVisibleParameterAnnotations of Annotation.t list list
  | `RuntimeInvisibleParameterAnnotations of Annotation.t list list
  | `AnnotationDefault of Annotation.element_value
  | `Unknown of utf8 * string ]

type for_class =
  [ `InnerClasses of ((Name.for_class option) * (Name.for_class option) * (utf8 option) * (AccessFlag.for_inner_class list)) list
  | `EnclosingMethod of Name.for_class * ((Name.for_method * Descriptor.for_method) option)
  | `Synthetic
  | `Signature of [`Class of Signature.class_signature]
  | `SourceFile of utf8
  | `SourceDebugExtension of utf8
  | `Deprecated
  | `RuntimeVisibleAnnotations of Annotation.t list
  | `RuntimeInvisibleAnnotations of Annotation.t list
  | `Unknown of utf8 * string ]

type element =
  | Class
  | Method
  | Field


(* Conversion functions *)

let get_class_name pool idx err =
  match ConstantPool.get_entry pool idx with
  | ConstantPool.Class idx' ->
      (match ConstantPool.get_entry pool idx' with
      | ConstantPool.UTF8 n -> Name.make_for_class_from_internal n
      | _ -> fail err)
  | _ -> fail err

let get_utf8 pool idx msg =
  match ConstantPool.get_entry pool idx with
  | ConstantPool.UTF8 v -> v
  | _ -> fail msg

let read_verification_type_info st pool =
  let tag = InputStream.read_u1 st in
  match tag with
  | 0 -> Top_variable_info
  | 1 -> Integer_variable_info
  | 2 -> Float_variable_info
  | 4 -> Long_variable_info
  | 3 -> Double_variable_info
  | 5 -> Null_variable_info
  | 6 -> Uninitialized_this_variable_info
  | 7 ->
      let idx = InputStream.read_u2 st in
      let d = match ConstantPool.get_entry pool idx with
      | ConstantPool.Class idx' ->
          (match ConstantPool.get_entry pool idx' with
          | ConstantPool.UTF8 v ->
              if uchar_eq opening_square_bracket (utf8_get v 0) then
                let t = Descriptor.java_type_of_utf8 v in `Array_type (Descriptor.filter_array Descriptor.Invalid_array_element_type t)
              else
                `Class_or_interface (Name.make_for_class_from_internal v)
          | _ -> fail Invalid_stack_map_verification_type)
      | _ -> fail Invalid_stack_map_verification_type in
      Object_variable_info d
  | 8 -> Uninitialized_variable_info (InputStream.read_u2 st)
  | _ -> fail Invalid_stack_map_verification_type

let write_verification_type_info st pool = function
  | Top_variable_info ->
      OutputStream.write_u1 st 0
  | Integer_variable_info ->
      OutputStream.write_u1 st 1
  | Float_variable_info ->
      OutputStream.write_u1 st 2
  | Long_variable_info ->
      OutputStream.write_u1 st 4
  | Double_variable_info ->
      OutputStream.write_u1 st 3
  | Null_variable_info ->
      OutputStream.write_u1 st 5
  | Uninitialized_this_variable_info ->
      OutputStream.write_u1 st 6
  | Object_variable_info d ->
      OutputStream.write_u1 st 7;
      let idx = match d with
      | `Class_or_interface u -> ConstantPool.add_class pool u
      | `Array_type t -> ConstantPool.add_array_class pool t in
      OutputStream.write_u2 st idx
  | Uninitialized_variable_info ofs ->
      OutputStream.write_u1 st 8;
      OutputStream.write_u2 st ofs

let read_stack_map_frame st pool prev_ofs =
  let map_ofs o = prev_ofs + o + 1 in
  let tag = InputStream.read_u1 st in
  match tag with
  | 247 ->
      let ofs = map_ofs (InputStream.read_u2 st) in
      let vti = read_verification_type_info st pool in
      (Same_locals_1_stack_item_frame (ofs, vti)), ofs
  | 248 ->
      let ofs = map_ofs (InputStream.read_u2 st) in
      (Chop_3_frame ofs), ofs
  | 249 ->
      let ofs = map_ofs (InputStream.read_u2 st) in
      (Chop_2_frame ofs), ofs
  | 250 ->
      let ofs = map_ofs (InputStream.read_u2 st) in
      (Chop_1_frame ofs), ofs
  | 251 ->
      let ofs = map_ofs (InputStream.read_u2 st) in
      (Same_frame ofs), ofs
  | 252 ->
      let ofs = map_ofs (InputStream.read_u2 st) in
      let t1 = read_verification_type_info st pool in
      (Append_1_frame (ofs, t1)), ofs
  | 253 ->
      let ofs = map_ofs (InputStream.read_u2 st) in
      let t1 = read_verification_type_info st pool in
      let t2 = read_verification_type_info st pool in
      (Append_2_frame (ofs, t1, t2)), ofs
  | 254 ->
      let ofs = map_ofs (InputStream.read_u2 st) in
      let t1 = read_verification_type_info st pool in
      let t2 = read_verification_type_info st pool in
      let t3 = read_verification_type_info st pool in
      (Append_3_frame (ofs, t1, t2, t3), ofs)
  | 255 ->
      let ofs = map_ofs (InputStream.read_u2 st) in
      let nb_locals = InputStream.read_u2 st in
      let locals = ref [] in
      for i = 1 to nb_locals do
        let l = read_verification_type_info st pool in
        locals := l :: !locals
      done;
      let nb_stack_items = InputStream.read_u2 st in
      let stack_items = ref [] in
      for i = 1 to nb_stack_items do
        let si = read_verification_type_info st pool in
        stack_items := si :: !stack_items
      done;
      (Full_frame (ofs, (List.rev !locals), (List.rev !stack_items))), ofs
  | _ when (tag >= 0) && (tag <= 63) ->
      let ofs = map_ofs tag in
      (Same_frame ofs), ofs
  | _ when (tag >= 64) && (tag <= 127) ->
      let ofs = map_ofs (tag - 64) in
      let vti = read_verification_type_info st pool in
      (Same_locals_1_stack_item_frame (ofs, vti)), ofs
  | _ ->
      fail Invalid_stack_map_frame

let write_stack_map_frame st pool prev_ofs x =
  let map_ofs o = o - prev_ofs - 1 in
  match x with
  | Same_frame o ->
      let ofs = map_ofs o in
      if (ofs >= 0) && (ofs <= 63) then
        (OutputStream.write_u1 st ofs;
         o)
      else
        (OutputStream.write_u1 st 251;
         OutputStream.write_u2 st ofs;
         o)
  | Same_locals_1_stack_item_frame (o, t) ->
      let ofs = map_ofs o in
      if (ofs >= 0) && (ofs <= 63) then
        (OutputStream.write_u1 st (64 + ofs);
         write_verification_type_info st pool t;
         o)
      else
        (OutputStream.write_u1 st 247;
         OutputStream.write_u2 st ofs;
         write_verification_type_info st pool t;
         o)
  | Chop_1_frame o ->
      let ofs = map_ofs o in
      OutputStream.write_u1 st 250;
      OutputStream.write_u2 st ofs;
      o
  | Chop_2_frame o ->
      let ofs = map_ofs o in
      OutputStream.write_u1 st 249;
      OutputStream.write_u2 st ofs;
      o
  | Chop_3_frame o ->
      let ofs = map_ofs o in
      OutputStream.write_u1 st 248;
      OutputStream.write_u2 st ofs;
      o
  | Append_1_frame (o, t1) ->
      let ofs = map_ofs o in
      OutputStream.write_u1 st 252;
      OutputStream.write_u2 st ofs;
      write_verification_type_info st pool t1;
      o
  | Append_2_frame (o, t1, t2) ->
      let ofs = map_ofs o in
      OutputStream.write_u1 st 253;
      OutputStream.write_u2 st ofs;
      write_verification_type_info st pool t1;
      write_verification_type_info st pool t2;
      o
  | Append_3_frame (o, t1, t2, t3) ->
      let ofs = map_ofs o in
      OutputStream.write_u1 st 254;
      OutputStream.write_u2 st ofs;
      write_verification_type_info st pool t1;
      write_verification_type_info st pool t2;
      write_verification_type_info st pool t3;
      o
  | Full_frame (o, locals, stack_items) ->
      let ofs = map_ofs o in
      let len_locals = List.length locals in
      let len_stack_items = List.length stack_items in
      if (len_locals <= 65535) && (len_stack_items <= 65535) then
        (OutputStream.write_u1 st 255;
         OutputStream.write_u2 st ofs;
         OutputStream.write_u2 st len_locals;
         List.iter (write_verification_type_info st pool) locals;
         OutputStream.write_u2 st len_stack_items;
         List.iter (write_verification_type_info st pool) stack_items;
         o)
      else
        fail Invalid_stack_map_frame

let check_code_attributes l =
  let map = function
    | (`LineNumberTable _) as x -> x
    | (`LocalVariableTable _) as x -> x
    | (`LocalVariableTypeTable _) as x -> x
    | (`Unknown _) as x -> x
    | (`StackMapTable _) as x -> x
    | #t -> fail Invalid_code_attribute in
  List.map map l

let check_field_attributes l =
  let fail_if b a x = if !b then fail (Defined_twice a) else b := true; x in
  let const = ref false in
  let synth = ref false in
  let sign = ref false in
  let depr = ref false in
  let v_ann = ref false in
  let i_ann = ref false in
  let map = function
    | (`ConstantValue _) as x -> fail_if const "ConstantValue" x
    | `Synthetic -> fail_if synth "Synthetic" `Synthetic
    | (`Signature (`Field _)) as x -> fail_if sign "Signature" x
    | `Deprecated -> fail_if depr "Deprecated" `Deprecated
    | (`RuntimeVisibleAnnotations _) as x -> fail_if v_ann "RuntimeVisibleAnnotations" x
    | (`RuntimeInvisibleAnnotations _) as x -> fail_if i_ann "RuntimeInvisibleAnnotations" x
    | (`Unknown _) as x -> x
    | #t -> fail Invalid_field_attribute in
  List.map map l

let check_method_attributes l =
  let fail_if b a x = if !b then fail (Defined_twice a) else b := true; x in
  let code = ref false in
  let exn = ref false in
  let synth = ref false in
  let sign = ref false in
  let depr = ref false in
  let v_ann = ref false in
  let i_ann = ref false in
  let v_p_ann = ref false in
  let i_p_ann = ref false in
  let ann_d = ref false in
  let map = function
    | (`Code _) as x -> fail_if code "Code" x
    | (`Exceptions _) as x -> fail_if exn "Exceptions" x
    | `Synthetic -> fail_if synth "Synthetic" `Synthetic
    | (`Signature (`Method _)) as x -> fail_if sign "Signature" x
    | `Deprecated -> fail_if depr "Deprecated" `Deprecated
    | (`RuntimeVisibleAnnotations _) as x -> fail_if v_ann "RuntimeVisibleAnnotations" x
    | (`RuntimeInvisibleAnnotations _) as x -> fail_if i_ann "RuntimeInvisibleAnnotations" x
    | (`RuntimeVisibleParameterAnnotations _) as x -> fail_if v_p_ann "RuntimeVisibleParameterAnnotations" x
    | (`RuntimeInvisibleParameterAnnotations _) as x -> fail_if i_p_ann "RuntimeInvisibleParameterAnnotations" x
    | (`AnnotationDefault _) as x -> fail_if ann_d "AnnotationDefault" x
    | (`Unknown _) as x -> x
    | #t -> fail Invalid_method_attribute in
  List.map map l

let check_class_attributes l =
  let fail_if b a x = if !b then fail (Defined_twice a) else b := true; x in
  let inner = ref false in
  let encl = ref false in
  let synth = ref false in
  let file = ref false in
  let debug = ref false in
  let sign = ref false in
  let depr = ref false in
  let v_ann = ref false in
  let i_ann = ref false in
  let map = function
    | (`InnerClasses _) as x -> fail_if inner "InnerClasses" x
    | (`EnclosingMethod _) as x -> fail_if encl "EnclosingMethod" x
    | `Synthetic -> fail_if synth "Synthetic" `Synthetic
    | (`SourceFile _) as x -> fail_if file "SourceFile" x
    | (`SourceDebugExtension _) as x -> fail_if debug "SourceDebugExtension" x
    | (`Signature (`Class _)) as x -> fail_if sign "Signature" x
    | `Deprecated -> fail_if depr "Deprecated" `Deprecated
    | (`RuntimeVisibleAnnotations _) as x -> fail_if v_ann "RuntimeVisibleAnnotations" x
    | (`RuntimeInvisibleAnnotations _) as x -> fail_if i_ann "RuntimeInvisibleAnnotations" x
    | (`Unknown _) as x -> x
    | #t -> fail Invalid_class_attribute in
  List.map map l

let rec decode element pool i =
  let st = InputStream.make_of_string i.data in
  let read_annotations () =
    let nb = InputStream.read_u2 st in
    let res = ref [] in
    for i = 1 to nb do
      let a = Annotation.read_info st in
      res := (Annotation.decode pool a) :: !res
    done;
    List.rev !res in
  let read_annotations_list () =
    let nb = InputStream.read_u1 st in
    let res = ref [] in
    for i = 1 to nb do
      let nb' = InputStream.read_u2 st in
      let local = ref [] in
      for j = 1 to nb' do
        let a = Annotation.read_info st in
        local := (Annotation.decode pool a) :: !local
      done;
      res := (List.rev !local) :: !res
    done;
    List.rev !res in
  let attr_name = get_utf8 pool i.name_index Invalid_attribute_name_value in
  if utf8_eq attr_name attr_constant_value then
    let const_index = InputStream.read_u2 st in
    match ConstantPool.get_entry pool const_index with
    | ConstantPool.Long (hi, lo) ->
        let v = Int64.logor (Int64.shift_left (Int64.of_int32 hi) 32) (Int64.of_int32 lo) in
        `ConstantValue (Long_value v)
    | ConstantPool.Float v ->
        `ConstantValue (Float_value (Int32.float_of_bits v))
    | ConstantPool.Double (hi, lo) ->
        let v = Int64.logor (Int64.shift_left (Int64.of_int32 hi) 32) (Int64.of_int32 lo) in
        `ConstantValue (Double_value (Int64.float_of_bits v))
    | ConstantPool.Integer v ->
        `ConstantValue (Integer_value v)
    | ConstantPool.String idx ->
        `ConstantValue (String_value (get_utf8 pool idx Invalid_constant_value))
    | _ -> fail Invalid_constant_value
  else if utf8_eq attr_name attr_code then
    let mx_stack = InputStream.read_u2 st in
    let mx_locals = InputStream.read_u2 st in
    let code_len' = InputStream.read_u4 st in
    let code_len =
      if code_len' < 65536L then
        Int64.to_int code_len'
      else
        fail Invalid_code_length in
    let code_content = InputStream.read_bytes st code_len in
    let exceptions_len = InputStream.read_u2 st in
    let exceptions = ref [] in
    for i = 1 to exceptions_len do
      let start_pc = InputStream.read_u2 st in
      let end_pc = InputStream.read_u2 st in
      let handler_pc = InputStream.read_u2 st in
      let catch_index = InputStream.read_u2 st in
      let catch_type =
        if catch_index <> 0 then
          Some (get_class_name pool catch_index Invalid_exception_name)
        else
          None in
      exceptions := (start_pc, end_pc, handler_pc, catch_type) :: !exceptions
    done;
    let attrib_count = InputStream.read_u2 st in
    let attrs = ref [] in
    for i = 1 to attrib_count do
      let a = read_info st in
      attrs := (decode element pool a) :: !attrs
    done;
    let code_stream = InputStream.make_of_string code_content in
    `Code { max_stack = mx_stack;
            max_locals = mx_locals;
            code = List.map (fun i -> Instruction.decode pool i) (ByteCode.read code_stream 0);
            exception_table = List.rev !exceptions;
            attributes = List.rev (check_code_attributes !attrs); }
  else if utf8_eq attr_name attr_exceptions then
    let exceptions_nb = InputStream.read_u2 st in
    let res = ref [] in
    for i = 1 to exceptions_nb do
      let idx = InputStream.read_u2 st in
      let name = get_class_name pool idx Invalid_exception in
      res := name :: !res
    done;
    `Exceptions (List.rev !res)
  else if utf8_eq attr_name attr_inner_classes then
    let nb = InputStream.read_u2 st in
    let res = ref [] in
    for i = 1 to nb do
      let inner_info_index = InputStream.read_u2 st in
      let outer_info_index = InputStream.read_u2 st in
      let inner_name_index = InputStream.read_u2 st in
      let inner_class_access_flag = InputStream.read_u2 st in
      let inner_class =
        if inner_info_index = 0 then
          None
        else
          Some (get_class_name pool inner_info_index Invalid_inner_class) in
      let outer_class =
        if outer_info_index = 0 then
          None
        else
          Some (get_class_name pool outer_info_index Invalid_outer_class) in
      let inner_name =
        if inner_name_index = 0 then
          None
        else
          Some (get_utf8 pool inner_name_index Invalid_inner_class) in
      let flags = AccessFlag.from_u2 false inner_class_access_flag in
      res := (inner_class, outer_class, inner_name, (AccessFlag.check_inner_class_flags flags)) :: !res
    done;
    `InnerClasses (List.rev !res)
  else if utf8_eq attr_name attr_enclosing_method then
    let class_index = InputStream.read_u2 st in
    let method_index = InputStream.read_u2 st in
    let class_name = get_class_name pool class_index Invalid_enclosing_method in
    let method_desc = if method_index <> 0 then
      try
        match pool.(method_index) with
        | ConstantPool.NameAndType (name, desc) ->
            Some ((Name.make_for_method (get_utf8 pool name Invalid_enclosing_method)),
                  (Descriptor.method_of_utf8 (get_utf8 pool desc Invalid_enclosing_method)))
        | _ -> fail Invalid_enclosing_method
      with _ -> fail Invalid_enclosing_method
    else
      None in
    `EnclosingMethod (class_name, method_desc)
  else if utf8_eq attr_name attr_synthetic then
    `Synthetic
  else if utf8_eq attr_name attr_signature then
    let signature_index = InputStream.read_u2 st in
    let s = get_utf8 pool signature_index Invalid_signature in
    let s' = (match element with
    | Class -> `Class (Signature.class_signature_of_utf8 s)
    | Method -> `Method (Signature.method_signature_of_utf8 s)
    | Field -> `Field (Signature.field_type_signature_of_utf8 s)) in
    `Signature s'
  else if utf8_eq attr_name attr_source_file then
    let sourcefile_index = InputStream.read_u2 st in
    `SourceFile (get_utf8 pool sourcefile_index Invalid_source_file)
  else if utf8_eq attr_name attr_source_debug_extension then
    let extension = utf8_of_modified (modified_of_bytes i.data) in
    `SourceDebugExtension extension
  else if utf8_eq attr_name attr_line_number_table then
    let nb = InputStream.read_u2 st in
    let res = ref [] in
    for i = 1 to nb do
      let start_pc = InputStream.read_u2 st in
      let line_number = InputStream.read_u2 st in
      res := (start_pc, line_number) :: !res
    done;
    `LineNumberTable (List.rev !res)
  else if utf8_eq attr_name attr_local_variable_table then
    let nb = InputStream.read_u2 st in
    let res = ref [] in
    for i = 1 to nb do
      let start_pc = InputStream.read_u2 st in
      let length = InputStream.read_u2 st in
      let name_index = InputStream.read_u2 st in
      let desc_index = InputStream.read_u2 st in
      let index = InputStream.read_u2 st in
      let name = get_utf8 pool name_index Invalid_local_variable_table in
      let desc = get_utf8 pool desc_index Invalid_local_variable_table in
      let field_desc = Descriptor.field_of_utf8 desc in
      if Name.is_valid_unqualified name then
        res := (start_pc, length, name, field_desc, index) :: !res
      else
        fail Invalid_local_variable_table
    done;
    `LocalVariableTable (List.rev !res)
  else if utf8_eq attr_name attr_local_variable_type_table then
    let nb = InputStream.read_u2 st in
    let res = ref [] in
    for i = 1 to nb do
      let start_pc = InputStream.read_u2 st in
      let length = InputStream.read_u2 st in
      let name_index = InputStream.read_u2 st in
      let sign_index = InputStream.read_u2 st in
      let index = InputStream.read_u2 st in
      let name = get_utf8 pool name_index Invalid_local_variable_type_table in
      let s = get_utf8 pool sign_index Invalid_local_variable_type_table in
      let sign = Signature.field_type_signature_of_utf8 s in
      if Name.is_valid_unqualified name then
        res := (start_pc, length, name, sign, index) :: !res
      else
        fail Invalid_local_variable_type_table
    done;
    `LocalVariableTypeTable (List.rev !res)
  else if utf8_eq attr_name attr_deprecated then
    `Deprecated
  else if utf8_eq attr_name attr_runtime_visible_annotations then
    `RuntimeVisibleAnnotations (read_annotations ())
  else if utf8_eq attr_name attr_runtime_invisible_annotations then
    `RuntimeInvisibleAnnotations (read_annotations ())
  else if utf8_eq attr_name attr_runtime_visible_parameter_annotations then
    `RuntimeVisibleParameterAnnotations (read_annotations_list ())
  else if utf8_eq attr_name attr_runtime_invisible_parameter_annotations then
    `RuntimeInvisibleParameterAnnotations (read_annotations_list ())
  else if utf8_eq attr_name attr_annotation_default then
    let eiv = Annotation.read_info_element_value st in
    `AnnotationDefault (Annotation.decode_element_value pool eiv)
  else if utf8_eq attr_name attr_stack_map_table then
    let nb = InputStream.read_u2 st in
    let res = ref [] in
    let ofs = ref (-1) in
    for i = 1 to nb do
      let e, o = read_stack_map_frame st pool !ofs in
      res := e :: !res;
      ofs := o
    done;
    `StackMapTable (List.rev !res)
  else
    `Unknown (attr_name, i.data)

let rec encode pool a =
  let checked_length l =
    let res = List.length l in
    if res < 65536 then
      res
    else
      fail Invalid_list_length in
  let buffer = Buffer.create 64 in
  let st = OutputStream.make_of_buffer buffer in
  let write_annotations l =
    let len = checked_length l in
    OutputStream.write_u2 st len;
    List.iter (fun a ->
      let a' = Annotation.encode pool a in
      Annotation.write_info st a') l in
  let write_annotations_list l =
    let len = checked_length l in
    OutputStream.write_u2 st len;
    List.iter (fun l' ->
      let len' = checked_length l' in
      OutputStream.write_u2 st len';
      List.iter (fun a ->
        let a' = Annotation.encode pool a in
        Annotation.write_info st a') l') l in
  let return n =
    let name_idx = ConstantPool.add_utf8 pool n in
    let content = Buffer.contents buffer in
    { name_index = name_idx;
      length = Int64.of_int (String.length content);
      data = content; } in
  match a with
  | `ConstantValue v ->
      (match v with
      | Long_value l ->
          let idx = ConstantPool.add_long pool l in
          OutputStream.write_u2 st idx;
          return attr_constant_value
      | Float_value f ->
          let idx = ConstantPool.add_float pool f in
          OutputStream.write_u2 st idx;
          return attr_constant_value
      | Double_value d ->
          let idx = ConstantPool.add_double pool d in
          OutputStream.write_u2 st idx;
          return attr_constant_value
      | Boolean_value b ->
          let idx = ConstantPool.add_integer pool (if b then 1l else 0l) in
          OutputStream.write_u2 st idx;
          return attr_constant_value
      | Byte_value v | Character_value v | Short_value v ->
          let idx = ConstantPool.add_integer pool (Int32.of_int v) in
          OutputStream.write_u2 st idx;
          return attr_constant_value
      | Integer_value i ->
          let idx = ConstantPool.add_integer pool i in
          OutputStream.write_u2 st idx;
          return attr_constant_value
      | String_value s ->
          let idx = ConstantPool.add_string pool s in
          OutputStream.write_u2 st idx;
          return attr_constant_value)
  | `Code c ->
      let code_content = List.map (fun i -> Instruction.encode pool i) c.code in
      let code_buffer = Buffer.create 16 in
      let code_stream = OutputStream.make_of_buffer code_buffer in
      ByteCode.write code_stream 0 code_content;
      OutputStream.close code_stream;
      let actual_code = Buffer.contents code_buffer in
      OutputStream.write_u2 st c.max_stack;
      OutputStream.write_u2 st c.max_locals;
      let code_length = String.length actual_code in
      if code_length >= 65536 then fail Invalid_code_length;
      OutputStream.write_u4 st (Int64.of_int code_length);
      OutputStream.write_bytes st actual_code;
      let len = checked_length c.exception_table in
      OutputStream.write_u2 st len;
      List.iter (fun (start_pc, end_pc, handler_pc, catch) ->
        let catch_idx = match catch with
        | Some exn_name -> ConstantPool.add_class pool exn_name
        | None -> 0 in
        OutputStream.write_u2 st start_pc;
        OutputStream.write_u2 st end_pc;
        OutputStream.write_u2 st handler_pc;
        OutputStream.write_u2 st catch_idx) c.exception_table;
      let len' = checked_length c.attributes in
      OutputStream.write_u2 st len';
      let sub_buffer = Buffer.create 16 in
      let sub_st = OutputStream.make_of_buffer sub_buffer in
      List.iter (fun a ->
        let a' = encode pool (a :> t) in
        write_info sub_st a') c.attributes;
      OutputStream.close sub_st;
      OutputStream.write_bytes st (Buffer.contents sub_buffer);
      return attr_code
  | `Exceptions l ->
      let len = checked_length l in
      OutputStream.write_u2 st len;
      List.iter (fun s ->
        let idx = ConstantPool.add_class pool s in
        OutputStream.write_u2 st idx) l;
      return attr_exceptions
  | `InnerClasses l ->
      let len = checked_length l in
      OutputStream.write_u2 st len;
      List.iter (fun (inner_class, outer_class, inner_name, flags) ->
        let inner_idx = match inner_class with
        | None -> 0
        | Some c -> ConstantPool.add_class pool c in
        let outer_idx = match outer_class with
        | None -> 0
        | Some c -> ConstantPool.add_class pool c in
        let name_idx = match inner_name with
        | None -> 0
        | Some c -> ConstantPool.add_utf8 pool c in
        let fl = AccessFlag.list_to_u2 (flags :> AccessFlag.t list) in
        OutputStream.write_u2 st inner_idx;
        OutputStream.write_u2 st outer_idx;
        OutputStream.write_u2 st name_idx;
        OutputStream.write_u2 st fl) l;
      return attr_inner_classes
  | `EnclosingMethod (class_name, meth_desc) ->
      let class_idx = ConstantPool.add_class pool class_name in
      let meth_idx = match meth_desc with
      | Some (n, d) ->
          ConstantPool.add_name_and_type pool
            (Name.utf8_for_method n)
            (Descriptor.utf8_of_method d)
      | None -> 0 in
      OutputStream.write_u2 st class_idx;
      OutputStream.write_u2 st meth_idx;
      return attr_enclosing_method
  | `Synthetic ->
      return attr_synthetic
  | `Signature s ->
      let s' = (match s with
      | `Class cs -> Signature.utf8_of_class_signature cs
      | `Method ms -> Signature.utf8_of_method_signature ms
      | `Field fs -> Signature.utf8_of_field_type_signature fs) in
      let idx = ConstantPool.add_utf8 pool s' in
      OutputStream.write_u2 st idx;
      return attr_signature
  | `SourceFile sf ->
      let idx = ConstantPool.add_utf8 pool sf in
      OutputStream.write_u2 st idx;
      return attr_source_file
  | `SourceDebugExtension sde ->
      let bytes = bytes_of_modified (modified_of_utf8 sde) in
      Buffer.add_string buffer bytes;
      return attr_source_debug_extension
  | `LineNumberTable l ->
      let len = checked_length l in
      OutputStream.write_u2 st len;
      List.iter (fun (start_pc, line_number) ->
        OutputStream.write_u2 st start_pc;
        OutputStream.write_u2 st line_number) l;
      return attr_line_number_table
  | `LocalVariableTable l ->
      let len = checked_length l in
      OutputStream.write_u2 st len;
      List.iter (fun (start_pc, length, name, desc, index) ->
        let name_index = ConstantPool.add_utf8 pool name in
        let desc_val = Descriptor.utf8_of_field desc in
        let desc_index = ConstantPool.add_utf8 pool desc_val in
        OutputStream.write_u2 st start_pc;
        OutputStream.write_u2 st length;
        OutputStream.write_u2 st name_index;
        OutputStream.write_u2 st desc_index;
        OutputStream.write_u2 st index) l;
      return attr_local_variable_table
  | `LocalVariableTypeTable l ->
      let len = checked_length l in
      OutputStream.write_u2 st len;
      List.iter (fun (start_pc, length, name, sign, index) ->
        let name_index = ConstantPool.add_utf8 pool name in
        let sign_val = Signature.utf8_of_field_type_signature sign in
        let sign_index = ConstantPool.add_utf8 pool sign_val in
        OutputStream.write_u2 st start_pc;
        OutputStream.write_u2 st length;
        OutputStream.write_u2 st name_index;
        OutputStream.write_u2 st sign_index;
        OutputStream.write_u2 st index) l;
      return attr_local_variable_type_table
  | `Deprecated ->
      return attr_deprecated
  | `RuntimeVisibleAnnotations l ->
      write_annotations l;
      return attr_runtime_visible_annotations
  | `RuntimeInvisibleAnnotations l ->
      write_annotations l;
      return attr_runtime_invisible_annotations
  | `RuntimeVisibleParameterAnnotations l ->
      write_annotations_list l;
      return attr_runtime_visible_parameter_annotations
  | `RuntimeInvisibleParameterAnnotations l ->
      write_annotations_list l;
      return attr_runtime_invisible_parameter_annotations
  | `AnnotationDefault ev ->
      let eiv = Annotation.encode_element_value pool ev in
      Annotation.write_info_element_value st eiv;
      return attr_annotation_default
  | `StackMapTable l ->
      let len = checked_length l in
      OutputStream.write_u2 st len;
      let ofs = ref (-1) in
      List.iter
        (fun x ->
          let o = write_stack_map_frame st pool !ofs x in
          ofs := o)
        l;
      return attr_stack_map_table
  | `Unknown (n, v) ->
      Buffer.add_string buffer v;
      return n

let rec version_bounds = function
  | `ConstantValue _ ->
      Version.make_bounds "'ConstantValue' attribute" Version.Java_1_0 None
  | `Code cv ->
      let code_bounds = Version.make_bounds "'Code' attribute" Version.Java_1_0 None in
      let instrs_bounds = List.map Instruction.version_bounds cv.code in
      let attrs_bounds = List.map version_bounds (cv.attributes :> t list) in
      Version.intersect_list (code_bounds :: (instrs_bounds @ attrs_bounds))
  | `Exceptions _ ->
      Version.make_bounds "'Exceptions' attribute" Version.Java_1_0 None
  | `InnerClasses _ ->
      Version.make_bounds "'InnerClasses' attribute" Version.Java_1_1 None
  | `EnclosingMethod _ ->
      Version.make_bounds "'EnclosingMethod' attribute" Version.Java_1_5 None
  | `Synthetic ->
      Version.make_bounds "'Synthetic' attribute" Version.Java_1_1 None
  | `Signature _ ->
      Version.make_bounds "'Signature' attribute" Version.Java_1_5 None
  | `SourceFile _ ->
      Version.make_bounds "'SourceFile' attribute" Version.Java_1_0 None
  | `SourceDebugExtension _ ->
      Version.make_bounds "'SourceDebugExtension' attribute" Version.Java_1_5 None
  | `LineNumberTable _ ->
      Version.make_bounds "'LineNumberTable' attribute" Version.Java_1_0 None
  | `LocalVariableTable _ ->
      Version.make_bounds "'LocalVariableTable' attribute" Version.Java_1_0 None
  | `LocalVariableTypeTable _ ->
      Version.make_bounds "'LocalVariableTypeTable' attribute" Version.Java_1_5 None
  | `Deprecated ->
      Version.make_bounds "'Deprecated' attribute" Version.Java_1_1 None
  | `RuntimeVisibleAnnotations _ ->
      Version.make_bounds "'RuntimeVisibleAnnotations' attribute" Version.Java_1_5 None
  | `RuntimeInvisibleAnnotations _ ->
      Version.make_bounds "'RuntimeInvisibleAnnotations' attribute" Version.Java_1_5 None
  | `RuntimeVisibleParameterAnnotations _ ->
      Version.make_bounds "'RuntimeVisibleParameterAnnotations' attribute" Version.Java_1_5 None
  | `RuntimeInvisibleParameterAnnotations _ ->
      Version.make_bounds "'RuntimeInvisibleParameterAnnotations' attribute" Version.Java_1_5 None
  | `AnnotationDefault _ ->
      Version.make_bounds "'AnnotationDefault' attribute" Version.Java_1_5 None
  | `StackMapTable _ ->
      Version.make_bounds "'StackMapTable' attribute" Version.Java_1_6 None
  | `Unknown _ ->
      Version.make_bounds "'Unknown' attribute" Version.Java_1_0 None
