(*
 * This file is part of Barista.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Barista is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Barista is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)

(* Types *)

type t =
  | From_string of string * int ref
  | From_buffer of Buffer.t * int ref
  | From_channel of in_channel


(* Exception *)

type error =
  | End_of_input_stream
  | Unable_to_read_data
  | Unable_to_close_stream
  | Data_is_too_large

exception Exception of error

let fail e = raise (Exception e)

let string_of_error = function
  | End_of_input_stream -> "end of input stream"
  | Unable_to_read_data -> "unable to read data"
  | Unable_to_close_stream -> "unable to close stream"
  | Data_is_too_large -> "data is too large"

let () =
  Printexc.register_printer
    (function
      | Exception e -> Some (string_of_error e)
      | _ -> None)


(* Constructors *)

let make_of_string s =
  From_string (s, ref 0)

let make_of_buffer b =
  From_buffer (b, ref 0)

let make_of_channel ch =
  set_binary_mode_in ch true;
  From_channel ch


(* Functions *)

let read_u1 st =
  match st with
  | From_string (str, ptr) ->
      let tmp = !ptr in
      if (tmp < String.length str) then
        let res = int_of_char str.[tmp] in
        ptr := tmp + 1;
        res
      else
        fail End_of_input_stream
  | From_buffer (buf, ptr) ->
      let tmp = !ptr in
      if (tmp < Buffer.length buf) then
        let res = int_of_char (Buffer.nth buf tmp) in
        ptr := tmp + 1;
        res
      else
        fail End_of_input_stream
  | From_channel ch ->
      try
        input_byte ch
      with
      | End_of_file -> fail End_of_input_stream
      | _ -> fail Unable_to_read_data

let read_u2 st =
  let a = read_u1 st in
  let b = read_u1 st in
  (a lsl 8) lor b

let read_u4 st =
  let a = Int64.of_int (read_u1 st) in
  let b = Int64.of_int (read_u1 st) in
  let c = Int64.of_int (read_u1 st) in
  let d = Int64.of_int (read_u1 st) in
  Int64.logor
    (Int64.shift_left a 24)
    (Int64.logor
       (Int64.shift_left b 16)
       (Int64.logor
          (Int64.shift_left c 8)
          d))

let read_s1 st =
  let x = read_u1 st in
  if x < 128 then x else x - 256

let read_s2 st =
  let x = read_u2 st in
  if x < 32768 then x else x - 65536

let read_s4 st =
  let a = Int32.of_int (read_u1 st) in
  let b = Int32.of_int (read_u1 st) in
  let c = Int32.of_int (read_u1 st) in
  let d = Int32.of_int (read_u1 st) in
  Int32.logor
    (Int32.shift_left a 24)
    (Int32.logor
       (Int32.shift_left b 16)
       (Int32.logor
          (Int32.shift_left c 8)
          d))

let read_bytes st nb =
  match st with
  | From_string (str, ptr) ->
      (try
        let tmp = !ptr in
        let res = String.sub str tmp nb in
        ptr := tmp + nb;
        res
      with Invalid_argument _ -> fail End_of_input_stream)
  | From_buffer (buf, ptr) ->
      (try
        let tmp = !ptr in
        let res = Buffer.sub buf tmp nb in
        ptr := tmp + nb;
        res
      with Invalid_argument _ -> fail End_of_input_stream)
  | From_channel ch ->
      (try
        let res = String.create nb in
        really_input ch res 0 nb;
        res
      with
      | Invalid_argument _ -> fail End_of_input_stream
      | End_of_file -> fail End_of_input_stream
      | _ -> fail Unable_to_read_data)

let read_bytes_into st nb dst idx =
  match st with
  | From_string (str, ptr) ->
      (try
        let tmp = !ptr in
        String.blit str tmp dst idx nb;
        ptr := tmp + nb
      with Invalid_argument _ -> fail End_of_input_stream)
  | From_buffer (buf, ptr) ->
      (try
        let tmp = !ptr in
        let res = Buffer.sub buf tmp nb in
        String.blit res 0 dst idx nb;
        ptr := tmp + nb
      with Invalid_argument _ -> fail End_of_input_stream)
  | From_channel ch ->
      (try
        really_input ch dst idx nb
      with
      | Invalid_argument _ -> fail End_of_input_stream
      | End_of_file -> fail End_of_input_stream
      | _ -> fail Unable_to_read_data)

let close = function
  | From_string (str, ptr) -> ptr := String.length str
  | From_buffer (buf, ptr) -> ptr := Buffer.length buf
  | From_channel ch -> try close_in ch with _ -> fail Unable_to_close_stream
