(*
 * This file is part of Barista.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Barista is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Barista is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)

open Utils


type action =
  | Print of string
  | Assemble of string
  | Disassemble of string
  | Flow of string

let version = "1.4"

let classpath = ref (try Sys.getenv "CLASSPATH" with Not_found -> ".")

let output = ref "."

let target = ref Version.Java_1_6

let compute_stacks = ref false

let actions = ref []

let api = ref false

let switches = [
  ("-version",
   Arg.Unit (fun () -> print_endline version; exit 0),
   " Print version and exit") ;

  ("-cp",
   Arg.String (fun s -> classpath := !classpath ^ ":" ^ s),
   "<colon-separated-list>  Add to classpath") ;

  ("-classpath",
   Arg.Set_string classpath,
   "<colon-separated-list>  Set classpath") ;

  ("-output",
   Arg.Set_string output,
   "<path>  Output path for generated class files") ;

  ("-target",
   Arg.String
     (function
       | "1.0" -> target := Version.Java_1_0
       | "1.1" -> target := Version.Java_1_1
       | "1.2" -> target := Version.Java_1_2
       | "1.3" -> target := Version.Java_1_3
       | "1.4" -> target := Version.Java_1_4
       | "1.5" -> target := Version.Java_1_5
       | "1.6" -> target := Version.Java_1_6
       | "1.7" -> target := Version.Java_1_7
       | x -> print_string x; print_endline " is not a valid version number"; exit 1),
   "<version>  Target version for generated class files") ;

  ("-compute-stacks",
   Arg.Set compute_stacks,
   " Computes stack elements (max_locals, max_stacks, and frames)");

  ("-print",
   Arg.String (fun s -> actions := (Print s) :: !actions),
   "<classname>  Print class") ;

  ("-asm",
   Arg.String (fun s -> actions := (Assemble s) :: !actions),
   "<file>  Assemble file") ;

  ("-dasm",
   Arg.String (fun s -> actions := (Disassemble s) :: !actions),
   "<classname>  Disassemble class") ;

  ("-flow",
   Arg.String (fun s -> actions := (Flow s) :: !actions),
   "<method>  Print control flow graph for method") ;

  ("-api",
   Arg.Set api,
   " Register API callbacks (internal use)")
]

let main () =
  Arg.parse
    switches
    (fun s ->
      if Filename.check_suffix s ".j" then
        actions := (Assemble s) :: !actions
      else
        actions := (Print s) :: !actions)
    "Barista: Java [dis]assembler";
  let headers =
    (List.fold_left
       (fun x a -> match a with
       | Print _ -> succ x
       | Assemble _ -> x
       | Disassemble _ -> succ x
       | Flow _ -> succ x)
       0
       !actions) >= 2 in
  let print_header s =
    if headers then
      let prefix = "--- " in
      let separator = " " in
      let size =
        List.fold_left
          (fun acc elem -> acc + (String.length elem))
          0
          [prefix; separator; s] in
      print_endline (prefix ^ s ^ separator ^ (String.make (max 0 (79 - size)) '-'))
    else
      () in
  let cp = ClassPath.make_of_string !classpath in
  let loader = ClassLoader.make cp in
  let output_path = Assembler.Path !output in
  List.iter
    (function
      | Print s ->
          print_header s;
          Printer.print cp (utf8_of_string s)
      | Assemble s ->
          let ic = open_in s in
          let name =
            Assembler.assemble
              ~version:!target
              ~compute_stacks:!compute_stacks
              ~class_loader:loader
              ic
              output_path in
          (try
            print_endline ((string_of_utf8 (Name.external_utf8_for_class name))
                           ^ " has been compiled")
          with _ -> ());
          (try close_in ic with _ -> ());
      | Disassemble s ->
          print_header s;
          Disassembler.disassemble cp (utf8_of_string s)
      | Flow s ->
          print_header s;
          let fail_param () = print_endline "invalid '-flow' parameter"; exit 1 in
          let class_name, method_name, (method_params, method_ret) =
            try
              let tokens = Lexer.tokens_of_line (utf8_of_string s) in
              match tokens with
              | [ Lexer.Method (cn, mn, md) ] -> cn, mn, md
              | _ -> fail_param ()
            with _ -> fail_param () in
          let clas = ClassLoader.find loader (Name.external_utf8_for_class class_name) in
          let meth =
            try
              let method_name = Name.utf8_for_method method_name in
              let is_constructor = utf8_eq method_name Consts.class_constructor in
              let is_initializer = utf8_eq method_name Consts.class_initializer in
              let rec eq_type x y =
                match (x, y) with
                | (`Class cn1), (`Class cn2) ->
                    Utils.utf8_eq
                      (Name.internal_utf8_for_class cn1)
                      (Name.internal_utf8_for_class cn2)
                | (`Array a1), (`Array a2) ->
                    eq_type (a1 :> Descriptor.java_type) (a2 :> Descriptor.java_type)
                | _ -> x = y in
              let eq_params l l' = 
                ((List.length l) = (List.length l'))
                  && (List.for_all2 eq_type (l :> Descriptor.java_type list) (l' :> Descriptor.java_type list)) in
              List.find
                (function
                  | Method.Regular (_, name, (params, ret), _) ->
                      (utf8_eq (Name.utf8_for_method name) method_name)
                        && (eq_params method_params params)
                        && (eq_type method_ret ret)
                  | Method.Constructor (_, params, _) ->
                      is_constructor
                        && (eq_params method_params params)
                        && (method_ret = `Void)
                | Method.Initializer _ ->
                    is_initializer && (method_params = []) && (method_ret = `Void))
                clas.ClassDefinition.methods
            with Not_found -> print_endline "method not found"; exit 1 in
          let rec extract_code = function
            | (`Code x) :: _ -> x
            | _ :: tl -> extract_code tl
            | [] -> print_endline "method has no code"; exit 1 in
          let code =
            extract_code
              (match meth with
              | Method.Regular (_, _, _, a) -> a
              | Method.Constructor (_, _, a) -> a
              | Method.Initializer (_, a) -> a) in
          let graph =
            ControlFlow.graph_of_instructions
              code.Attribute.code
              code.Attribute.exception_table in
          let dot =
            ControlFlow.dot_of_graph
              (Printf.sprintf "[offset %ld]")
              (fun _ -> "")
              graph in
          print_endline dot)
    (List.rev !actions);
  ClassPath.close cp

let () =
  try
    main ();
    if !api then
      Api.register_callbacks ()
    else
      exit 0
  with
  | Assembler.Exception (n, e) ->
      Printf.printf "line %d: %s\n" n (Assembler.string_of_error e);
      exit 1
  | ClassPath.Exception e ->
    print_endline (ClassPath.string_of_error e);
    exit 1
  | ClassLoader.Exception e ->
    print_endline (ClassLoader.string_of_error e);
    exit 1
  | e ->
    print_endline (Printexc.to_string e);
    exit 1
