(*
 * This file is part of Barista.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Barista is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Barista is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)

open Utils
open Consts


(* Low-level form *)

type info = {
    access_flags : u2;
    name_index : u2;
    descriptor_index : u2;
    attributes_count : u2;
    attributes : Attribute.info array;
  }


(* Exception *)

type error =
  | Invalid_name
  | Invalid_name_value
  | Invalid_descriptor_value

exception Exception of error

let fail e = raise (Exception e)

let string_of_error = function
  | Invalid_name -> "invalid name"
  | Invalid_name_value -> "invalid name value"
  | Invalid_descriptor_value -> "invalid descriptor value"

let () =
  Printexc.register_printer
    (function
      | Exception e -> Some (string_of_error e)
      | _ -> None)


(* I/O functions *)

let read_info st =
  let flags = InputStream.read_u2 st in
  let name = InputStream.read_u2 st in
  let desc = InputStream.read_u2 st in
  let att_count = InputStream.read_u2 st in
  let atts = Array.init att_count (fun _ -> Attribute.read_info st) in
  { access_flags = flags;
    name_index = name;
    descriptor_index = desc;
    attributes_count = att_count;
    attributes = atts; }

let write_info st i =
  OutputStream.write_u2 st i.access_flags;
  OutputStream.write_u2 st i.name_index;
  OutputStream.write_u2 st i.descriptor_index;
  OutputStream.write_u2 st i.attributes_count;
  Array.iter (Attribute.write_info st) i.attributes


(* High-level form *)

type t =
  | Regular of (AccessFlag.for_method list) * Name.for_method * Descriptor.for_method * (Attribute.for_method list)
  | Constructor of (AccessFlag.for_constructor list) * (Descriptor.for_parameter list) * (Attribute.for_method list)
  | Initializer of bool * (Attribute.for_method list)


(* Conversion functions *)

let decode itf pool i =
  let name = match ConstantPool.get_entry pool i.name_index with
  | ConstantPool.UTF8 n ->
      if Name.is_valid_for_method n then
        n
      else
        fail Invalid_name
  | _ -> fail Invalid_name_value in
  let desc = match ConstantPool.get_entry pool i.descriptor_index with
   | ConstantPool.UTF8 d -> Descriptor.method_of_utf8 d
   | _ -> fail Invalid_descriptor_value in
   let attrs = Attribute.check_method_attributes (List.map (Attribute.decode Attribute.Method pool) (Array.to_list i.attributes)) in
    if (utf8_eq class_initializer name) then
      let flags = AccessFlag.from_u2 true i.access_flags in
      Initializer ((List.mem `Strict flags), attrs)
    else if (utf8_eq class_constructor name) then
      let flags = AccessFlag.check_constructor_flags (AccessFlag.from_u2 true i.access_flags) in
      Constructor (flags, (fst desc), attrs)
    else
      let flags = AccessFlag.check_method_flags itf (AccessFlag.from_u2 true i.access_flags) in
      Regular (flags, (Name.make_for_method name), desc, attrs)

let encode pool m =
  let flags, name, desc, attrs = match m with
  | Regular (f, n, d, a) -> (f, n, d, a)
  | Constructor (f, d, a) -> ((f :> AccessFlag.for_method list), (Name.make_for_method class_constructor), (d, `Void), a)
  | Initializer (s, a) -> ((if s then [`Strict] else []), (Name.make_for_method class_initializer), ([], `Void), a) in
  let acc_flags = AccessFlag.list_to_u2 (flags :> AccessFlag.t list) in
  let name_idx = ConstantPool.add_utf8 pool (Name.utf8_for_method name) in
  let desc_utf8 = Descriptor.utf8_of_method desc in
  let desc_idx = ConstantPool.add_utf8 pool desc_utf8 in
  { access_flags = acc_flags;
    name_index = name_idx;
    descriptor_index = desc_idx;
    attributes_count = List.length attrs;
    attributes = Array.of_list (List.map (Attribute.encode pool) (attrs :> Attribute.t list)); }
