(*
 * This file is part of Barista.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Barista is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Barista is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)

open Utils
open Consts


(* Utility functions *)

let replace_dot_with_slash =
  utf8_replace dot slash

let replace_slash_with_dot s =
  utf8_replace dollar dot (utf8_replace slash dot s)

let is_valid_unqualified s =
  ((utf8_length s) > 0)
    && not (utf8_contains dot s)
    && not (utf8_contains semi_colon s)
    && not (utf8_contains opening_square_bracket s)
    && not (utf8_contains slash s)

let is_valid_for_method s =
  (utf8_eq s class_constructor)
|| (utf8_eq s class_initializer)
|| ((is_valid_unqualified s)
      && not (utf8_contains lower_than s)
      && not (utf8_contains greater_than s))


(* Exception *)

type error =
  | Invalid_class_name
  | Invalid_field_name
  | Invalid_method_name

exception Exception of error

let fail e = raise (Exception e)

let string_of_error = function
  | Invalid_class_name -> "invalid class name"
  | Invalid_field_name -> "invalid field name"
  | Invalid_method_name -> "invalid method name"

let () =
  Printexc.register_printer
    (function
      | Exception e -> Some (string_of_error e)
      | _ -> None)


(* Name types and conversion functions *)

type for_class = (utf8 list) * (utf8 list)

type for_field = utf8

type for_method = utf8

let make_for_class ch s =
  let check x =
    if is_valid_unqualified x then
      x
    else
      fail Invalid_class_name in
  try
    let idx = utf8_index_from s 0 dollar in
    let cls = utf8_substring s 0 (pred idx) in
    let inner = utf8_substring s (succ idx) (pred (utf8_length s)) in
    (List.map check (utf8_split ch cls)),
    (List.map check (utf8_split dollar inner))
  with Not_found -> (List.map check (utf8_split ch s)), []

let make_for_class_from_internal = make_for_class slash

let make_for_class_from_external = make_for_class dot

let make_for_field s =
  if is_valid_unqualified s then
    s
  else
    fail Invalid_field_name

let make_for_method s =
  if is_valid_for_method s then
    s
  else
    fail Invalid_method_name

let utf8_for_class sep1 sep2 c =
  let sep1' = make_utf8 [sep1] in
  let sep2' = make_utf8 [sep2] in
  let cls, inner = c in
  if inner = [] then
    utf8_concat_sep sep1' cls
  else
    (utf8_concat_sep sep1' cls) ++ sep2' ++ (utf8_concat_sep sep2' inner)

let printable_utf8_for_class = utf8_for_class dot dot

let external_utf8_for_class = utf8_for_class dot dollar

let internal_utf8_for_class = utf8_for_class slash dollar

let utf8_for_field f = f

let utf8_for_method m = m
