/*
 * This file is part of Barista.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Barista is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Barista is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.io.FileOutputStream;
import java.util.Collections;
import java.util.List;
import static java.util.Arrays.asList;

import fr.x9c.barista.api.*;
import static fr.x9c.barista.api.Instruction.*;
import static tests.Common.*;

public final class Source {

    private static final FieldRef systOut =
	new FieldRef("java.lang.System",
		     "out",
		     new Descriptor.Class("java.io.PrintStream"));

    public static void main(final String[] args) throws Throwable {
	final Field x = new Field(asList(AccessFlag.Private,
					 AccessFlag.Final),
				  "x",
				  Descriptor.INT,
				  Collections.<Attribute>emptyList());
	final Field y = new Field(asList(AccessFlag.Private,
					 AccessFlag.Final),
				  "y",
				  Descriptor.INT,
				  Collections.<Attribute>emptyList());
	final Field prefix = new Field(asList(AccessFlag.Private,
					     AccessFlag.Static),
				       "prefix",
				       new Descriptor.Class("java.lang.String"),
				       Collections.<Attribute>singletonList(new Attribute.ConstantValue(new Attribute.StringValue("prefix"))));
	final List<Instruction> clinitInstructions =
	    asList(new LDC_STRING("state: "),
		   new PUTSTATIC(new FieldRef("pack.Test",
					      "prefix",
					      new Descriptor.Class("java.lang.String"))),
		   RETURN);
	final List<Instruction> initInstructions =
	    asList(ALOAD_0,
		   new INVOKESPECIAL(new MethodRef("java.lang.Object",
						   "<init>",
						   Descriptor.VOID,
						   descs())),
		   ALOAD_0,
		   ILOAD_1,
		   new PUTFIELD (new FieldRef("pack.Test",
					      "x",
					      Descriptor.INT)),
		   ALOAD_0,
		   ILOAD_2,
		   new PUTFIELD (new FieldRef("pack.Test",
					      "y",
					      Descriptor.INT)),
		   RETURN);
	final List<Instruction> setInstructions =
	    asList(ALOAD_0,
		   ILOAD_1,
		   new PUTFIELD (new FieldRef("pack.Test",
					      "x",
					      Descriptor.INT)),
		   ALOAD_0,
		   ILOAD_2,
		   new PUTFIELD (new FieldRef("pack.Test",
					      "y",
					      Descriptor.INT)),
		   RETURN);
	final List<Instruction> toStringInstructions =
	    asList(new NEW(new TypeName("java.lang.StringBuilder")),
		   DUP,
		   new INVOKESPECIAL(new MethodRef("java.lang.StringBuilder",
						   "<init>",
						   Descriptor.VOID,
						   descs())),
		   new GETSTATIC(new FieldRef("pack.Test",
					      "prefix",
					      new Descriptor.Class("java.lang.String"))),
		   new INVOKEVIRTUAL(new MethodRef("java.lang.StringBuilder",
						   "append",
						   new Descriptor.Class("java.lang.StringBuilder"),
						   descs(new Descriptor.Class("java.lang.String")))),
		   new LDC_STRING("x="),
		   new INVOKEVIRTUAL(new MethodRef("java.lang.StringBuilder",
						   "append",
						   new Descriptor.Class("java.lang.StringBuilder"),
						   descs(new Descriptor.Class("java.lang.String")))),
		   ALOAD_0,
		   new GETFIELD(new FieldRef("pack.Test",
					     "x",
					     Descriptor.INT)),
		   new INVOKEVIRTUAL(new MethodRef("java.lang.StringBuilder",
						   "append",
						   new Descriptor.Class("java.lang.StringBuilder"),
						   descs(Descriptor.INT))),
		   new LDC_STRING(", y="),
		   new INVOKEVIRTUAL(new MethodRef("java.lang.StringBuilder",
						   "append",
						   new Descriptor.Class("java.lang.StringBuilder"),
						   descs(new Descriptor.Class("java.lang.String")))),
		   ALOAD_0,
		   new GETFIELD(new FieldRef("pack.Test",
					     "y",
					     Descriptor.INT)),
		   new INVOKEVIRTUAL(new MethodRef("java.lang.StringBuilder",
						   "append",
						   new Descriptor.Class("java.lang.StringBuilder"),
						   descs(Descriptor.INT))),
		   new INVOKEVIRTUAL(new MethodRef("java.lang.StringBuilder",
						   "toString",
						   new Descriptor.Class("java.lang.String"),
						   descs())),
		   ARETURN);
	final List<Instruction> instructions =
	    asList(new NEW(new TypeName("pack.Test")),
		   DUP,
		   ICONST_3,
		   ICONST_5,
		   new INVOKESPECIAL(new MethodRef("pack.Test",
						   "<init>",
						   Descriptor.VOID,
						   descs(Descriptor.INT,
							 Descriptor.INT))),
		   DUP,
		   DUP,
		   DUP2,
		   POP2,
		   new GETSTATIC(systOut),
		   SWAP,
		   new INVOKEVIRTUAL(new MethodRef("java.io.PrintStream",
						   "println",
						   Descriptor.VOID,
						   descs(new Descriptor.Class("java.lang.Object")))),
		   new SIPUSH((short) -3),
		   new BIPUSH((byte) 7),
		   new INVOKEVIRTUAL(new MethodRef("pack.Test",
						   "set",
						   Descriptor.VOID,
						   descs(Descriptor.INT,
							 Descriptor.INT))),
		   new GETSTATIC(systOut),
		   SWAP,
		   new INVOKEVIRTUAL(new MethodRef("java.io.PrintStream",
						   "println",
						   Descriptor.VOID,
						   descs(new Descriptor.Class("java.lang.Object")))),
		   RETURN);
	final Method clinit = compileInitializer(clinitInstructions);
	final Method init = compileConstructor(asList(AccessFlag.Public),
					       descs(Descriptor.INT,
						     Descriptor.INT),
					       4,
					       4,
					       Collections.<Attribute.ExceptionTableElement>emptyList(),
					       Collections.<Attribute>emptyList(),
					       initInstructions);
	final Method set =
	    compileMethod(asList(AccessFlag.Private),
			  "set",
			  Descriptor.VOID,
			  descs(Descriptor.INT, Descriptor.INT),
			  4,
			  4,
			  Collections.<Attribute.ExceptionTableElement>emptyList(),
			  Collections.<Attribute>emptyList(),
			  Collections.<Attribute>emptyList(),
			  setInstructions);
	final Method toString =
	    compileMethod(asList(AccessFlag.Public),
			  "toString",
			  new Descriptor.Class("java.lang.String"),
			  descs(),
			  4,
			  4,
			  Collections.<Attribute.ExceptionTableElement>emptyList(),
			  Collections.<Attribute>emptyList(),
			  Collections.<Attribute>emptyList(),
			  toStringInstructions);
	final Method main = compileMethod(asList(AccessFlag.Public, AccessFlag.Static),
					  "main",
					  Descriptor.VOID,
					  descs(new Descriptor.Array(new Descriptor.Class("java.lang.String"))),
					  8,
					  4,
					  Collections.<Attribute.ExceptionTableElement>emptyList(),
					  Collections.<Attribute>emptyList(),
					  Collections.<Attribute>emptyList(),
					  instructions);
	final ClassDefinition cd = compileClass(null,
						null,
						null,
						asList("java.io.Serializable"),
						asList(x, y, prefix),
						null,
						asList(clinit,
						       init,
						       set,
						       toString,
						       main));
	final byte[] bc = ByteCode.encode(cd);
	FileOutputStream out = new FileOutputStream("pack/Test.class");
	out.write(bc);
	out.flush();
    }

}
